/* 
  Magnum 
  Copyright (C) 2002-2004 Kaya Memisoglu. All Rights Reserved.

  This software is provided 'as-is', without any express or implied
  warranty.  In no event will the authors be held liable for any damages
  arising from the use of this software.

  Permission is granted to anyone to use this software for any purpose,
  including commercial applications, and to alter it and redistribute it
  freely, under the terms in LICENSE.TXT.

  Kaya Memisoglu  (k.memisoglu@dimajix.de)
*/

#ifndef __GEOMETRY_FRUSTUM_INL
#define __GEOMETRY_FRUSTUM_INL

namespace magnum { namespace geometry {

/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
template<class T, int dim> 
inline Frustum<T,dim>::Frustum() {
	m_Offset.null();
	m_Origin.null();
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
template<class T, int dim> 
inline void Frustum<T,dim>::setOrigin(const Vector<T,dim>& v) {
	m_Origin = v;
	m_NearOrigin = m_Origin + m_Near*m_Direction;
	m_FarOrigin = m_Origin + m_Far*m_Direction;
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
template<class T, int dim> 
inline const Vector<T,dim>& Frustum<T,dim>::getOrigin() const {
	return m_Origin;
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
template<class T, int dim>
inline void	Frustum<T,dim>::setAxes(const Vector<T,dim>* e) {
	for (unsigned int i = 0; i < dim-1; i++) {
		m_Axes[i] = e[i];
		m_Axes2[i] = e[i] * e[i];
		m_OffsetAxes2[i] = m_Offset[i] * m_Axes2[i];
	}
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
template<class T, int dim>
inline const Vector<T,dim>* Frustum<T,dim>::getAxes() const {
	return m_Axes;
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
template<class T, int dim>
inline void	Frustum<T,dim>::setAxis(unsigned int i, const Vector<T,dim>& a) {
	ASSERT(i < dim-1);
	m_Axes[i] = a;
	m_Axes2[i] = a*a;
	m_OffsetAxes2[i] = m_Offset[i] * m_Axes2[i];
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
template<class T, int dim>
inline const Vector<T,dim>& Frustum<T,dim>::getAxis(unsigned int i) const {
	ASSERT(i < dim-1);
	return m_Axes[i];
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
template<class T, int dim> 
inline void Frustum<T,dim>::setDirection(const Vector<T,dim>& v) {
	m_Direction = v;
	m_Direction2 = v*v;
	m_NearDirection2 = m_Direction2 * m_Near;
	m_FarDirection2 = m_Direction2 * m_Far;
	m_NearOrigin = m_Origin + m_Near*m_Direction;
	m_FarOrigin = m_Origin + m_Far*m_Direction;
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
template<class T, int dim> 
inline const Vector<T,dim>& Frustum<T,dim>::getDirection() const {
	return m_Direction;
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
template<class T, int dim> 
inline void Frustum<T,dim>::setOffset(const Vector<T,dim-1>& v) {
	m_Offset = v;
	for (unsigned int i = 0; i < dim-1; i++)
		m_OffsetAxes2[i] = m_Offset[i] * m_Axes2[i];
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
template<class T, int dim> 
inline const Vector<T,dim-1>& Frustum<T,dim>::getOffset() const {
	return m_Offset;
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
template<class T, int dim> 
inline void Frustum<T,dim>::setNear(const T& n) {
	m_Near = n;
	m_FarByNear = m_Far / m_Near;
	m_NearDirection2 = m_Direction2 * m_Near;
	m_NearOrigin = m_Origin + m_Near*m_Direction;
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
template<class T, int dim> 
inline const T& Frustum<T,dim>::getNear() const {
	return m_Near;
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
template<class T, int dim> 
inline void Frustum<T,dim>::setFar(const T& f) {
	m_Far = f;
	m_FarByNear = m_Far / m_Near;
	m_FarDirection2 = m_Direction2 * m_Far;
	m_FarOrigin = m_Origin + m_Far*m_Direction;
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
template<class T, int dim> 
inline const T& Frustum<T,dim>::getFar() const {
	return m_Far;
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
template<class T, int dim> 
bool Frustum<T,dim>::isEmpty() const {
	if (m_Near == m_Far)
		return true;
	for (unsigned int i = 0; i < dim-1; i++) {
		if (m_Axes2[i] <= T(0))
			return true;
	}

	return false;
}
					

/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
template<class T, int dim> 
void Frustum<T,dim>::getVertices(Vector<T,dim>* v) const {
	unsigned int i,j;
	Vector<T,dim> center;

	// Get near plane vertices	
	center = m_NearOrigin;
	for (i = 0; i < dim; i++)
		center += m_Offset[j] * m_Axes[j];
	for (i = 0; i < (1 << (dim-1)); i++) {
		*v = center;
		for (j = 0; j < dim-1; j++) {
			if (i & (1 << j))
				*v += m_Axes[j];
			else
				*v -= m_Axes[j];
		}
		v++
	}

	// Get far plane vertices
	center = m_FarOrigin;
	for (i = 0; i < dim; i++)
		center += m_FarByNear * m_Offset[j] * m_Axes[j];
	for (i = 0; i < (1 << (dim-1)); i++) {
		*v = center;
		for (j = 0; j < dim-1; j++) {
			if (i & (1 << j))
				*v += m_FarByNear*m_Axes[j];
			else
				*v -= m_FarByNear*m_Axes[j];
		}
		v++
	}
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
template<class T, int dim> 
bool Frustum<T,dim>::isInside(const Vector<T,dim>& v) const {
	// Check against near and far plane
	T t = (v - m_Origin) * m_Direction;
	if ((t <= m_NearDirection2) || (t >= m_FarDirection2))
		return false;
		
	// Project v onto projection hyperplane
	Vector<T,dim> p = (m_NearDirection2 / t) * (vec - m_Origin);
	T l;
	if (dim == 2) {
		l = m_Axes[0]*p - m_OffsetAxes2[0];
		if ((l < -m_Axes2[0]) || (l > m_Axes2[0]))
			return false;
	} else if (dim == 3) {
		l = m_Axes[0]*p - m_OffsetAxes2[0];
		if ((l < -m_Axes2[0]) || (l > m_Axes2[0]))
			return false;
		l = m_Axes[1]*p - m_OffsetAxes2[1];
		if ((l < -m_Axes2[1]) || (l > m_Axes2[1]))
			return false;
	} else {
		for (unsigned int i = 0; i < dim-1; i++) {
			l = m_Axes[i]*p - m_OffsetAxes2[i];
			if ((l < -m_Axes2[i]) || (l > m_Axes2[i]))
				return false;
		}
	}

	return true;
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
template<class T, int dim> 
T Frustum<T,dim>::getDistance(const Vector<T,dim>& v) const {
	// Project v onto projection hyperplane
	T t = (v - m_Origin) * m_Direction;
	Vector<T,dim> p = (m_NearDirection2 / abs(t)) * (v - m_Origin);

	// Find nearest point on hyperplane
	Vector<T,dim> best;
	best.null();
	for (unsigned int i = 0; i < dim-1; i++) {
		T l = (p*m_Axes[i])/m_Axes2[i] - m_Offset[i]; 
		if (l < T(-1)) 
			l = T(-1);
		else if (l > T(1))
			l = T(1);
		best += (l + m_Offset[i]) * m_Axes[i];
	}

	// Check distances
	if (t < m_NearDirection2) {
		best += m_NearOrigin;
	} else if (t > m_FarDirection2) {
		// TODO: Bug, this is not really correct. We should
		// find a point perpendicular tp the closest plane
		best = m_FarOrigin + m_FarByNear * best;
	} else {
		t /= m_Direction2;
		best = m_Origin + t/m_Near*best + t * m_Direction;
	}
	
	return (best - v).length();
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
template<class T, int dim> 
inline bool Frustum<T,dim>::intersect(const Sphere<T,dim>& sphere) const {
	return getDistance(sphere.getOrigin()) <= sphere.getRadius();
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
template<class T, int dim> 
bool Frustum<T,dim>::intersect(const Cube<T,dim>& cube) const {
	unsigned int i,j;
	Vector<T,dim> p = cube.getOrigin();
	T radius = cube.getRadius();
	T t,l,m,tabs;
	T mi, ma;

	// Project Frustum onto cube axes
	for (i = 0; i < dim; i++) {
		// Project Frustum onto ith axes of cube
		t = m_Offset[0]*m_Axes[0][i] + abs(m_Axes[0][i]);
		for (j = 1; j < dim-1; j++)
			t += m_Offset[j]*m_Axes[j][i] + abs(m_Axes[j][i]);

		l = m_FarByNear * t;
		mi = min(m_NearOrigin[i] - t, m_FarOrigin[i] - l);
		ma = max(m_NearOrigin[i] + t, m_FarOrigin[i] + l);
		
		if ((ma <= p[i] - radius) || (mi >= p[i] + radius))
			return false;
	}

	// Project Cube onto Direction Vector
	p -= m_Origin;
	t = m_Direction * p;
	l = m_Direction.sumabs() * radius;
	if ((t - l >= m_FarDirection2) || (t+l <= m_NearDirection2))
		return false;
		
	// t is projection scale factor
	t = m_NearDirection2 / t;
	tabs = abs(t);
		
	// Project Cube onto projection hyperplane
	for (i = 0; i < dim-1; i++) {
		m = t * (m_Axes[i] * p);
		l = tabs * m_Axes[i].sumabs() * radius;
		if ((m+l <= m_OffsetAxes2[i] - m_Axes2[i]) || (m-l >= m_OffsetAxes2[i] + m_Axes2[i]))
			return false;
	}
	
	// Use combinations
	if (dim >= 3) {
	
	}
	
	return true;
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
template<class T, int dim> 
bool Frustum<T,dim>::intersect(const AlignedBox<T,dim>& box) const {
	unsigned int i,j;
	Vector<T,dim> bmin = box.getMinimum();
	Vector<T,dim> bmax = box.getMaximum();
	T t,l,m,tabs;
	T mi, ma;

	// Project Frustum onto aligned box axes
	for (i = 0; i < dim; i++) {
		// Project Frustum onto ith axes of box
		t = m_Offset[0]*m_Axes[0][i] + abs(m_Axes[0][i]);
		for (j = 1; j < dim-1; j++)
			t += m_Offset[j]*m_Axes[j][i] + abs(m_Axes[j][i]);

		l = m_FarByNear * t;
		mi = min(m_NearOrigin[i] - t, m_FarOrigin[i] - l);
		ma = max(m_NearOrigin[i] + t, m_FarOrigin[i] + l);
		
		if ((ma <= bmin[i]) || (mi >= bmax[i]))
			return false;
	}

	// Project box onto Direction Vector
	Vector<T,dim> bdiff = bmax - bmin;
	Vector<T,dim> bcent = bmax + bmin - m_Origin;
	t = m_Direction * bcent;
	l = abs(m_Direction[0] * bdiff[0]);
	for (j = 1; j < dim; j++)
		l += abs(m_Direction[j] * bdiff[j]);
	if ((t - l >= 2*m_FarDirection2) || (t+l <= 2*m_NearDirection2))
		return false;
		
	// t is projection scale factor
	t = m_NearDirection2 / t;
	tabs = abs(t);
		
	// Project box onto projection hyperplane
	for (i = 0; i < dim-1; i++) {
		m = t * (m_Axes[i] * bcent);
		l = abs(m_Axes[i][0] * bdiff[0]);
		for (j = 1; j < dim; j++)
			l += abs(m_Axes[i][j] * bdiff[j]);
		l *= tabs;
		if ((m+l <= 2 * (m_OffsetAxes2[i] - m_Axes2[i])) || (m-l >= 2*(m_OffsetAxes2[i] + m_Axes2[i])))
			return false;
	}
	
	return true;
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
template<class T, int dim> 
bool Frustum<T,dim>::intersect(const Box<T,dim>& ) const {
	return true;
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
template<class T, int dim> 
bool Frustum<T,dim>::intersect(const Bounds<T,dim>& ) const {
	return true;
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
template<class T, int dim> 
Containment Frustum<T,dim>::cull(const Cube<T,dim>& ) const {
	return Subset;
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
template<class T, int dim> 
Containment Frustum<T,dim>::cull(const Bounds<T,dim>& ) const {
	return Subset;
}

}; };

#endif
