/* 
  Magnum 
  Copyright (C) 2002-2004 Kaya Memisoglu. All Rights Reserved.

  This software is provided 'as-is', without any express or implied
  warranty.  In no event will the authors be held liable for any damages
  arising from the use of this software.

  Permission is granted to anyone to use this software for any purpose,
  including commercial applications, and to alter it and redistribute it
  freely, under the terms in LICENSE.TXT.

  Kaya Memisoglu  (k.memisoglu@dimajix.de)
*/

#ifndef __GEOMETRY_SPACE_INL
#define __GEOMETRY_SPACE_INL

namespace magnum { namespace geometry {
using magnum::math::abs;
using magnum::math::max;

/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
template<class T, int dim, class Base> 
inline Placeable<T,dim,Base>::Placeable() {
	m_Space = NULL;
	m_Prev = m_Next = NULL;
	m_Index = ~0;
	m_Node = NULL;
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
template<class T, int dim, class Base> 
Placeable<T,dim,Base>::~Placeable() {
	if (m_Space == NULL)
		return;

	ASSERT(m_Node != NULL);

	// Is item a normal leaf or a direct item
	if (m_Index == ~0) {
		if (m_Prev != NULL)
			m_Prev->m_Next = m_Next;
		else
			m_Node->m_Leafs = m_Next;
		if (m_Next != NULL)
			m_Next->m_Prev = m_Prev;
	} else {
		ASSERT(m_Node->m_ChildItem[m_Index] == this);
		m_Node->m_ChildItem[m_Index] = NULL;
		m_Node->m_ChildTypes &= ~(1 << m_Index);
	}

	// Tidy up node if needed
	m_Space->collapseNode(m_Node);

	// Adjust item counter
	m_Space->m_Size--;

	m_Space = NULL;
	m_Node = NULL;
	m_Next = m_Prev = NULL;
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
template<class T, int dim, class Base> 
inline Space<T,dim,Base>* Placeable<T,dim,Base>::getSpace() const {
	return m_Space;
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
template<class T, int dim, class Base> 
inline const Bounds<T,dim>& Placeable<T,dim,Base>::getBounds() const {
	return m_Bounds;
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
template<class T, int dim, class Base> 
inline void	Placeable<T,dim,Base>::setBounds(const Bounds<T,dim>& b) {
	m_Bounds = b;
	m_Bounds.toAlignedBox(&m_Box);
	if (m_Space != NULL)
		m_Space->update((Base*)this);
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
template<class T, int dim, class Base> 
Placeable<T,dim,Base>* Placeable<T,dim,Base>::getNextNode() const {
	Node* node = m_Node;
	unsigned int index = m_Index;
	Placeable<T,dim,Base>* item;

	if (index == ~0) {
		item = item->m_Next;
		if (item != NULL)
			return item;
		index = 0;
	}
	
	for (;;) {
		for (;index < (1 << dim); index++) {
			if (node->m_ChildNode[index] == NULL) 
				continue;
			if (node->m_ChildTypes & (1 << index))
				return node->m_ChildItem[index];

			// We found a nonempty subnode
			node = node->m_ChildTypes[index];

			// First process internal leafs
			if (node->m_Leafs != NULL)
				return node->m_Leafs;
					
			// Process subnodes
			index = 0;
		}
		
		// We found no other subnode
		index = node->m_Index;
		node = node->m_Parent;
		if (node == NULL)
			return NULL;
	}
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
template<class T, int dim, class Item> 
Space<T,dim,Item>::Space() {
	m_Size = 0;
	m_MaxDepth = 8;
	m_Root.setOrigin(magnum::math::Vector<T,dim>::Null);
	m_Root.setRadius(T(0));
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
template<class T, int dim, class Item> 
Space<T,dim,Item>::~Space() {
	clear();
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
template<class T, int dim, class Item> 
typename Space<T,dim,Item>::Iterator Space<T,dim,Item>::insert(Item* item) {
	ASSERT(item != NULL);
	Node* node;
	unsigned int index;

	// Add reference
	item->refer();
	m_Size++;

	Placeable<T,dim,Item>* pitem = (Placeable<T,dim,Item>*)item;
	ASSERT(pitem->m_Space == NULL);
	ASSERT(pitem->m_Node == NULL);

	// Set space
	pitem->m_Space = this;

	// Get bounding information
	Vector<T,dim> imin = pitem->m_Box.getMinimum();
	Vector<T,dim> imax = pitem->m_Box.getMaximum();
	Vector<T,dim> icenter = (imin + imax)/T(2);
	T isize = (imax - imin).maxabs();

	// Get minimal radius needed
	T neededRadius = max(imin.maxabs(), imax.maxabs());
	if (neededRadius > m_Root.getRadius())
		expandTree(neededRadius);

	// Insert item
	node = &m_Root;
	for(;;) {
		ASSERT(node != NULL);

		// Is item so big that we have to insert it directly
		// into the node?
		if (node->getRadius() <= isize) {
			pitem->m_Prev = NULL;
			pitem->m_Next = node->m_Leafs;
			if (pitem->m_Next != NULL)
				pitem->m_Next->m_Prev = pitem;
			pitem->m_Node = node;
			pitem->m_Index = ~0;
			node->m_Leafs = pitem;
			return Iterator(pitem);
		}

		// Get correct index for insertion
		index = getIndex(icenter, node->getOrigin());

		// Check if we already have an occupied child pointer
		if (node->m_ChildNode[index] == NULL) {
			// Empty subnode => direct insertion!
			node->m_ChildTypes |= (1 << index);
			node->m_ChildItem[index] = pitem;
			pitem->m_Node = node;
			pitem->m_Index = index;
			return Iterator(pitem);
		} 
		
		// Is node already occupied by an direct item?
		if (node->m_ChildTypes & (1 << index))
			expandChild(node, index);

		// Go to next subnode
		node = node->m_ChildNode[index];
		ASSERT(node->m_Index == index);
	}
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
template<class T, int dim, class Item> 
void Space<T,dim,Item>::remove(Item* item) {
	ASSERT(item != NULL);

	Placeable<T,dim,Item>* pitem = (Placeable<T,dim,Item>*)item;
	ASSERT(pitem->m_Space == this);

	Node* node = pitem->m_Node;
	ASSERT(node != NULL);

	unsigned int index = pitem->m_Index;

	// Is item a normal leaf or a direct item
	if (index == ~0) {
		if (pitem->m_Prev != NULL)
			pitem->m_Prev->m_Next = pitem->m_Next;
		else
			node->m_Leafs = pitem->m_Next;
		if (pitem->m_Next != NULL)
			pitem->m_Next->m_Prev = pitem->m_Prev;
	} else {
		ASSERT(node->m_ChildItem[index] == pitem);
		node->m_ChildItem[index] = NULL;
		node->m_ChildTypes &= ~(1 << index);
	}

	// Clear any fields in item
	pitem->m_Space = NULL;
	pitem->m_Node = NULL;
	pitem->m_Prev = pitem->m_Next = NULL;

	// Release item and adjust item counter
	item->release();
	m_Size--;

	// Tidy up node if needed
	collapseNode(node);
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
template<class T, int dim, class Item> 
inline void Space<T,dim,Item>::remove(const typename Space<T,dim,Item>::Iterator& iter) {
	remove((Item*)iter);
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
template<class T, int dim, class Item> 
void Space<T,dim,Item>::update(Item* item) {
	ASSERT(item != NULL);

	Placeable<T,dim,Item>* pitem = (Placeable<T,dim,Item>*)item;
	ASSERT(pitem->m_Space == this);
	
	ASSERT(false);
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
template<class T, int dim, class Item> 
bool Space<T,dim,Item>::contains(const Item* item) const {
	ASSERT(item != NULL);
	return ((Placeable<T,dim,Item>*)item)->m_Space == this;
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
template<class T, int dim, class Item> 
void Space<T,dim,Item>::clear() {
	Placeable<T,dim,Item>* pitem;
	Item* item;
	Node* node = &m_Root;
	unsigned int index = 0;

	for (;;) {
		// Did we find another subnode in last pass?
		if (index == (1 << dim)) {
			// We found no other subnode
			Node* oldnode = node;

			// Release any leaf items
			while(node->m_Leafs != NULL) {
				pitem = node->m_Leafs;
				ASSERT(pitem->m_Node == node);
				ASSERT(pitem->m_Index == ~0);

				node->m_Leafs = pitem->m_Next;

				// Clear any data memebers
				pitem->m_Space = NULL;
				pitem->m_Node = NULL;
				pitem->m_Next = pitem->m_Prev = NULL;

				// Release item
				item = (Item*)pitem;
				item->release();
				m_Size--;
			}

			// Go to parent node
			index = node->m_Index;
			node = node->m_Parent;

			// Do we really have a node, or are we ready?
			if (node == NULL) {
				ASSERT(m_Size == 0);
				return;
			}
			operator delete(oldnode, m_Allocator);

			// Next index in parent node!
			index++;
		}

		for (;index < (1 << dim); index++) {
			// Empty subnode => continue;
			if (node->m_ChildNode[index] == NULL) 
				continue;

			// Direct item => release
			if (node->m_ChildTypes & (1 << index)) {
				pitem = node->m_ChildItem[index];
				ASSERT(pitem->m_Node == node);
				ASSERT(pitem->m_Index == index);

				// Clear any data memebers
				pitem->m_Space = NULL;
				pitem->m_Node = NULL;
				pitem->m_Next = pitem->m_Prev = NULL;

				// Release item
				item = (Item*)pitem;
				item->release();
				m_Size--;

				continue;
			}

			// We found a nonempty subnode
			node = node->m_ChildNode[index];
			break;
		}
	}
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
template<class T, int dim, class Item> 
inline const Cube<T,dim>& Space<T,dim,Item>::getBounds() const {
	return m_Root.m_Cube;
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
template<class T, int dim, class Item> 
inline unsigned int Space<T,dim,Item>::getSize() const {
	return m_Size;
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
template<class T, int dim, class Item> 
inline bool Space<T,dim,Item>::isEmpty() const {
	return (m_Size == 0);
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
template<class T, int dim, class Item> 
inline void Space<T,dim,Item>::setMaxDepth(unsigned int d) {
	m_MaxDepth = d;
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
template<class T, int dim, class Item> 
inline unsigned int Space<T,dim,Item>::getMaxDepth() const {
	return m_MaxDepth;
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
template<class T, int dim, class Item> 
inline typename Space<T,dim,Item>::Iterator Space<T,dim,Item>::begin() const {
	return Iterator(getFirstItem());
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
template<class T, int dim, class Item> 
inline typename Space<T,dim,Item>::Iterator Space<T,dim,Item>::at(const Item* item) const {
	return Iterator((Placeable<T,dim,Item>*)item);
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
template<class T, int dim, class Item> 
inline typename Space<T,dim,Item>::Iterator Space<T,dim,Item>::end() const {
	return Iterator(NULL);
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
template<class T, int dim, class Item> 
Placeable<T,dim,Item>* Space<T,dim,Item>::getFirstItem() const {
	if (m_Root == NULL)
		return NULL;
	if (m_Root->m_Leafs != NULL)
		return m_Root->m_Leafs;

	Node* node = m_Root;
	unsigned int index = 0;

	for (;;) {
		if (node == NULL)
			return NULL;

		for (;index < (1 << dim); index++) {
			if (node->m_ChildNode[index] == NULL) 
				continue;
			if (node->m_ChildTypes & (1 << index))
				return node->m_ChildItem[index];

			// We found a nonempty subnode
			node = node->m_ChildTypes[index];

			// First process internal leafs
			if (node->m_Leafs != NULL)
				return node->m_Leafs;
					
			// Process subnodes
			index = 0;
		}
		
		// We found no other subnode
		index = node->m_Index;
		node = node->m_Parent;
	}
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
template<class T, int dim, class Item> 
void Space<T,dim,Item>::createTree(Item* item) {
	Placeable<T,dim,Item>* pitem = item;
	ASSERT(pitem->m_Space == NULL);
	ASSERT(pitem->m_Node == NULL);

	// Set space
	pitem->m_Space = this;

	// Get bounding information
	Vector<T,dim> imin = pitem->m_Box.getMinimum();
	Vector<T,dim> imax = pitem->m_Box.getMaximum();
	Vector<T,dim> icenter = (imin + imax)/T(2);
	
	m_RootSize = max(imin.maxabs(), imax.maxabs()) * T(2);
	m_Size = 1;

	// CReate root-node
	m_Root = new(m_Allocator) Node();

	// Insert item
	unsigned int index;
	index = getIndex(icenter, magnum::math::Vector<T,dim>::Null);
	m_Root->m_ChildTypes |= (1 << index);
	m_Root->m_ChildItem[index] = pitem;
	m_Root->setOrigin(magnum::math::Vector<T,dim>::Null);
	m_Root->setRadius(m_RootSize);
	pitem->m_Node = m_Root;
	pitem->m_Index = index;
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
template<class T, int dim, class Item> 
void Space<T,dim,Item>::expandTree(T neededRadius) {
	unsigned int i,j;
	Node* node;
	T rootsize = m_Root.getRadius();
	
	if (rootsize == T(0)) {
		rootsize = T(1);
		while(rootsize < neededRadius)
			rootsize *= T(2);
		m_Root.setRadius(rootsize);
		return;
	}

	// Do we need to expand the tree?
	while (neededRadius > rootsize) {
		for (i = 0; i < (1 << dim); i++) {
			// Check for empty subnode
			if (m_Root.m_ChildNode[i] == NULL) {
				continue;
			}

			// Check for direct item
			if (m_Root.m_ChildTypes & (1 << i)) {
				continue;
			} 

			// We have a real subnode!
			j = i ^ ((1 << dim) - 1);
			node = new(m_Allocator) Node();
			node->m_Cube.setRadius(rootsize);
			node->m_Cube.setOrigin(getCenter(magnum::math::Vector<T,dim>::Null, rootsize, i));
			node->m_ChildNode[j] = m_Root.m_ChildNode[i];
			node->m_ChildNode[j]->m_Parent = node;
			node->m_ChildNode[j]->m_Index = j;
			node->m_Parent = &m_Root;
			node->m_Index = i;
			m_Root.m_ChildNode[i] = node;
		}
		rootsize *= T(2);
		m_Root.setRadius(rootsize);
	}
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
template<class T, int dim, class Item> 
void Space<T,dim,Item>::expandChild(typename Space<T,dim,Item>::Node* node, unsigned int index) {
	ASSERT(node != NULL);
	ASSERT(node->m_ChildTypes & (1 << index));

	Placeable<T,dim,Item>* item = node->m_ChildItem[index];
	ASSERT(item != NULL);
	
	// Calculate center of new subnode
	T radius = node->getRadius() / T(2);
	Vector<T,dim> origin = getCenter(node->getOrigin(), radius, index);

	// We have to create a real new subnode
	Node* newnode = new(m_Allocator) Node();
	node->m_ChildTypes &= ~(1 << index);
	node->m_ChildNode[index] = newnode;

	// Initialize new node
	newnode->m_Parent = node;
	newnode->m_Index = index;
	newnode->m_Cube.setRadius(radius);
	newnode->m_Cube.setOrigin(origin);

	// Set new parent of item
	item->m_Node = newnode;

	// Check size: Shall we insert as a leaf or as a child?
	T isize = (item->m_Box.getMaximum() - item->m_Box.getMinimum()).maxabs();
	if (isize >= radius) {
		newnode->m_Leafs = item;
		item->m_Index = ~0;
	} else {
		index = getIndex(item->m_Box.getOrigin(), origin);
		newnode->m_ChildTypes = (1 << index);
		newnode->m_ChildItem[index] = item;
		item->m_Index = index;
	}
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
template<class T, int dim, class Item> 
void Space<T,dim,Item>::collapseNode(typename Space<T,dim,Item>::Node* node) {
	// Check if this node is still needed...
	if (node->m_Leafs != NULL)
		return;

	unsigned int i, size = 0;
	unsigned int index;
	for (i = 0; i < (1 << dim); i++) {
		if (node->m_ChildItem[i] == NULL)
			continue;
		if (!node->m_ChildTypes & (1 << i))
			return;
		index = i;
		size++;
	}

	// Do we have more than one child?
	if (size > 1)
		return;

	// Get parent node
	Node* parent = node->m_Parent;
	unsigned int pidx = node->m_Index;
	if (parent == NULL)
		return;

	Placeable<T,dim,Item>* pitem = node->m_ChildItem[index];

	parent->m_ChildTypes |= (1 << pidx);
	parent->m_ChildItem[pidx] = pitem;
	pitem->m_Index = pidx;
	pitem->m_Node = parent;
	operator delete(node, m_Allocator);
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
template<class T, int dim, class Item> 
inline unsigned int Space<T,dim,Item>::getIndex(const Vector<T,dim>& icenter, const Vector<T,dim>& scenter) const {
	unsigned int index = 0;
	for (unsigned int i = 0; i < dim; i++) {
		if (icenter[i] > scenter[i])
			index |= (1 << i);
	}
	return index;
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
template<class T, int dim, class Item> 
inline Vector<T,dim> Space<T,dim,Item>::getCenter(const Vector<T,dim>& oldcenter, T radius, unsigned int index) const {
	Vector<T,dim> origin = oldcenter;
	for (unsigned int i = 0; i < dim; i++) {
		if (index & (1 << i)) 
			origin[i] += radius;
		else
			origin[i] -= radius;
	}

	return origin;
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
template<class T, int dim, class Item>
template<class V, class F> 
void Space<T,dim,Item>::intersect(const V& volume, F& function) const {
	const Node* node = &m_Root;
	const Node* subnode;
	Containment cullcode;
	Placeable<T,dim,Item>* pitem;
	int index = 0;
	
	for (;;) {
		for (;index < (1 << dim); index++) {
		recurse:
			subnode = node->m_ChildNode[index];
			
			// Is there a nonemtpy entry?
			if (subnode == NULL) 
				continue;
				
			// Direct item?
			if (node->m_ChildTypes & (1 << index)) {
				pitem = (Placeable<T,dim,Item>*)(void *)subnode;
				if (volume.intersect(pitem->m_Bounds))
					function((Item*)pitem);
				continue;
			}

			// We found a nonempty subnode
			ASSERT(subnode->m_Index == index);
			cullcode = volume.cull(subnode->m_Cube);
			if (cullcode == Independent)
				continue;
				
			if (cullcode == Subset)	{
				intersectAll(function, subnode);
			} else {
				node = subnode;
				index = 0;
				goto recurse;
			}
		}

		// Check if we need to walk up
		do {
			// Traverse all Leaf items
			pitem = node->m_Leafs;
			while(pitem) {
				if (volume.intersect(pitem->m_Bounds))
					function((Item*)pitem);
				pitem = pitem->m_Next;
			}

			// We found no other subnode
			index = node->m_Index + 1;
			node = node->m_Parent;
			if (node == NULL)
				return;
		} while (index == (1 << dim));
	}
} 


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
template<class T, int dim, class Item>
template<class F> 
void Space<T,dim,Item>::intersect(const Vector<T,dim>& point, F& function) const {
	const Node* node = &m_Root;
	Placeable<T,dim,Item>* pitem;
	unsigned int index = 0;

	for (;;) {
		// Traverse all Leaf items
		pitem = node->m_Leafs;
		while(pitem) {
			if (pitem->m_Bounds.isInside(point))
				function((Item*)pitem);
			pitem = pitem->m_Next;
		}

		// Get correct index for insertion
		index = getIndex(point, node->getOrigin());

		// Is there a nonemtpy entry?
		if (node->m_ChildNode[index] == NULL) 
			return;
				
		// Direct item?
		if (node->m_ChildTypes & (1 << index)) {
			pitem = node->m_ChildItem[index];
			if (pitem->m_Bounds.isInside(point))
				function((Item*)pitem);
			return;
		}

		// We found a nonempty subnode
		node = node->m_ChildNode[index];
	}
} 


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
template<class T, int dim, class Item>
template<class F>
void Space<T,dim,Item>::intersectAll(F& function, const typename Space<T,dim,Item>::Node* root) const {
	const Node* node = root;
	const Node* subnode;
	Placeable<T,dim,Item>* pitem;
	int index = 0;

	for (;;) {
		for (;index < (1 << dim); index++) {
		recurse:
			subnode = node->m_ChildNode[index];
			
			// Is there a nonemtpy entry?
			if (subnode == NULL) 
				continue;
				
			// Direct item?
			if (node->m_ChildTypes & (1 << index)) {
				function((Item*)(Placeable<T,dim,Item>*)(void *)subnode);
				continue;
			}

			// We found a nonempty subnode
			ASSERT(subnode->m_Index == index);
			ASSERT(subnode->m_Parent == node);
			node = subnode;
			index = 0;
			goto recurse;
		}

		// Check if we need to walk up
		do {
			// Traverse all Leaf items
			pitem = node->m_Leafs;
			while(pitem) {
				function((Item*)pitem);
				pitem = pitem->m_Next;
			}

			// We found no other subnode
			if (node == root)
				return;
			
			index = node->m_Index + 1;
			node = node->m_Parent;
		} while (index == (1 << dim));
	}
} 



/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
template<class T, int dim, class Item> 
inline Space<T,dim,Item>::Node::Node() {
	m_Parent = NULL;
	m_Leafs = NULL;
	m_ChildTypes = 0;
	for (unsigned int i = 0; i < (1 << dim); i++)
		m_ChildNode[i] = NULL;
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
template<class T, int dim, class Item> 
inline void Space<T,dim,Item>::Node::setOrigin(const Vector<T,dim>& o) {
	m_Cube.setOrigin(o);
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
template<class T, int dim, class Item> 
inline const Vector<T,dim>&	Space<T,dim,Item>::Node::getOrigin() const {
	return m_Cube.getOrigin();
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
template<class T, int dim, class Item> 
inline void Space<T,dim,Item>::Node::setRadius(const T& r) {
	m_Cube.setRadius(r);
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
template<class T, int dim, class Item> 
inline const T& Space<T,dim,Item>::Node::getRadius() const {
	return m_Cube.getRadius();
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
template<class T, int dim, class Item> 
inline SpaceIterator<T,dim,Item>::SpaceIterator(Placeable<T,dim,Item>* item) {
	m_Item = item;
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
template<class T, int dim, class Item> 
inline SpaceIterator<T,dim,Item>::SpaceIterator() {
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
template<class T, int dim, class Item> 
inline Item* SpaceIterator<T,dim,Item>::operator->() const {
	return (Item*)m_Item;
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
template<class T, int dim, class Item> 
inline Item& SpaceIterator<T,dim,Item>::operator*() const {
	return *(Item*)m_Item;
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
template<class T, int dim, class Item> 
inline SpaceIterator<T,dim,Item>::operator Item*() const {
	return (Item*)m_Item;
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
template<class T, int dim, class Item> 
inline SpaceIterator<T,dim,Item>& SpaceIterator<T,dim,Item>::operator++() {
	m_Item = m_Item->getNextNode();
	return *this;
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
template<class T, int dim, class Item> 
inline SpaceIterator<T,dim,Item> SpaceIterator<T,dim,Item>::operator++(int ) {
	Placeable<T,dim,base>* item = m_Item;
	m_Item = m_Item->getNextNode();
	return SpaceIterator(item);
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
template<class T, int dim, class Item> 
inline SpaceIterator<T,dim,Item>& SpaceIterator<T,dim,Item>::operator--();


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
template<class T, int dim, class Item> 
inline SpaceIterator<T,dim,Item> SpaceIterator<T,dim,Item>::operator--(int );

/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
template<class T, int dim, class Item> 
inline bool SpaceIterator<T,dim,Item>::operator==(const SpaceIterator<T,dim,Item>& iter) const {
	return m_Item == iter.m_Item;
}

/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
template<class T, int dim, class Item> 
inline bool SpaceIterator<T,dim,Item>::operator!=(const SpaceIterator<T,dim,Item>& iter) const {
	return m_Item != iter.m_Item;
}

/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
template<class T, int dim, class Item> 
inline bool SpaceIterator<T,dim,Item>::operator<(const SpaceIterator<T,dim,Item>& iter) const {
	return m_Item < iter.m_Item;
}

/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
template<class T, int dim, class Item> 
inline bool	SpaceIterator<T,dim,Item>::operator>(const SpaceIterator<T,dim,Item>& iter) const {
	return m_Item > iter.m_Item;
}

/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
template<class T, int dim, class Item> 
inline bool SpaceIterator<T,dim,Item>::operator<=(const SpaceIterator<T,dim,Item>& iter) const {
	return m_Item <= iter.m_Item;
}

/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
template<class T, int dim, class Item> 
inline bool SpaceIterator<T,dim,Item>::operator>=(const SpaceIterator<T,dim,Item>& iter) const {
	return m_Item >= iter.m_Item;
}


}; };

#endif
