/* 
  Magnum 
  Copyright (C) 2002-2004 Kaya Memisoglu. All Rights Reserved.

  This software is provided 'as-is', without any express or implied
  warranty.  In no event will the authors be held liable for any damages
  arising from the use of this software.

  Permission is granted to anyone to use this software for any purpose,
  including commercial applications, and to alter it and redistribute it
  freely, under the terms in LICENSE.TXT.

  Kaya Memisoglu  (k.memisoglu@dimajix.de)
*/

#ifndef __GEOMETRY_TRANSFORM_H
#define __GEOMETRY_TRANSFORM_H

#include "math.matrix.h"
#include "math.vector.h"
#include "math.vector2.h"
#include "math.vector3.h"
#include "math.vector4.h"

namespace magnum { namespace geometry {
using magnum::math::Matrix;
using magnum::math::Vector;
using magnum::math::Vector2;
using magnum::math::Vector3;
using magnum::math::Vector4;

/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
template<class T, int dim> class Transform {
public:
	enum Type {
		Id,
		Scale,
		Translation,
		Orthogonal,
		Rigid,
		Congruent,
		Affine,
	};

private:
	Matrix<T,dim+1>		m_Matrix;
	Type				m_Type;
	T					m_Scale;

public:
	static const Transform<T,dim>	Identity;

public:
	inline Transform();
	inline explicit Transform(const Matrix<T,dim+1>& );
	inline explicit Transform(const Matrix<T,dim>& );
	inline explicit Transform(const Vector<T,dim>& );
	inline Transform(const Matrix<T,dim>& ,const Vector<T,dim>& );

	inline Type					getType() const;
	inline T					getScale() const;
								
	inline Transform<T,dim>		operator-() const;
								
	inline Transform<T,dim>		operator+(const Transform<T,dim>& ) const;
	inline Transform<T,dim>		operator+(const Matrix<T,dim+1>& ) const;
	inline Transform<T,dim>		operator+(const Vector<T,dim>& ) const;
	inline Transform<T,dim>		operator-(const Transform<T,dim>& ) const;
	inline Transform<T,dim>		operator-(const Matrix<T,dim+1>& ) const;
	inline Transform<T,dim>		operator-(const Vector<T,dim>& ) const;
	inline Transform<T,dim>		operator*(const Transform<T,dim>& ) const;
	inline Transform<T,dim>		operator*(const Matrix<T,dim+1>& ) const;
	inline Transform<T,dim>		operator*(const Matrix<T,dim>& ) const;
	inline Transform<T,dim>		operator*(const T& ) const;
	inline Transform<T,dim>		operator/(const T& ) const;
								
	inline Vector<T,dim>		operator*(const Vector<T,dim>& v) const;
	inline Vector<T,dim+1>		operator*(const Vector<T,dim+1>& v) const;


	inline const Transform<T,dim>&	operator+=(const Transform<T,dim>& );
	inline const Transform<T,dim>&	operator+=(const Matrix<T,dim+1>& );
	inline const Transform<T,dim>&	operator+=(const Vector<T,dim>& );
	inline const Transform<T,dim>&	operator-=(const Transform<T,dim>& );
	inline const Transform<T,dim>&	operator-=(const Matrix<T,dim+1>& );
	inline const Transform<T,dim>&	operator-=(const Vector<T,dim>& );
	inline const Transform<T,dim>&	operator*=(const Transform<T,dim>& );
	inline const Transform<T,dim>&	operator*=(const Matrix<T,dim+1>& );
	inline const Transform<T,dim>&	operator*=(const Matrix<T,dim>& );
	inline const Transform<T,dim>&	operator*=(const T& );
	inline const Transform<T,dim>&	operator/=(const T& );

	inline Transform<T,dim>		getInverse() const;
								
	inline Transform<T,dim>&	invert();
	inline Transform<T,dim>&	identity();
								
	inline void					setTranslation(const Vector<T,dim>& );
	inline Vector<T,dim>		getTranslation() const;
								
	inline void					setRotation(const Matrix<T,dim>& );
	inline Matrix<T,dim>		getRotation() const;

	inline void					setTransform(const Matrix<T,dim+1>& );
	inline const Matrix<T,dim+1>& getTransform() const;
	
	inline void					setAxis(unsigned int ,const Vector<T,dim>& );
	inline Vector<T,dim>		getAxis(unsigned int ) const;
								
	inline Transform<T,dim>&	translate(const Vector<T,dim>& );
	inline Transform<T,dim>&	rotate(const T&, const Vector<T,dim>& );
	inline Transform<T,dim>&	scale(const Vector<T,dim>& );
	inline Transform<T,dim>&	scale(const T& );
								
	inline						operator Matrix<T,dim+1>&();
	inline						operator const Matrix<T,dim+1>&() const;
};


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
template<class T> class Transform2 : public Transform<T,2> {
public:
	inline Transform2() {};
	inline Transform2(const Transform<T,2>& t) : Transform<T,2>(t) {};
	inline explicit Transform2(const Matrix<T,3>& m) : Transform<T,2>(m) {};
	inline explicit Transform2(const Matrix<T,2>& m) : Transform<T,2>(m) {};
	inline explicit Transform2(const Vector<T,2>& v) : Transform<T,2>(v) {};
	inline explicit Transform2(const Vector2<T>& v)  : Transform<T,2>((Vector<T,2>)v) {};
	inline Transform2(const Matrix<T,2>& m,const Vector<T,2>& v) : Transform<T,2>(m,v) {};
	inline Transform2(const Matrix<T,2>& m,const Vector2<T>& v)  : Transform<T,2>(m,(Vector<T,2>)v) {};
};

/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
template<class T> class Transform3 : public Transform<T,3> {
public:
	inline Transform3() {};
	inline Transform3(const Transform<T,3>& t) : Transform<T,3>(t) {};
	inline explicit Transform3(const Matrix<T,4>& m) : Transform<T,3>(m) {};
	inline explicit Transform3(const Matrix<T,3>& m) : Transform<T,3>(m) {};
	inline explicit Transform3(const Vector<T,3>& v) : Transform<T,3>(v) {};
	inline explicit Transform3(const Vector3<T>& v)  : Transform<T,3>((Vector<T,3>)v) {};
	inline Transform3(const Matrix<T,3>& m,const Vector<T,3>& v) : Transform<T,3>(m,v) {};
	inline Transform3(const Matrix<T,3>& m,const Vector3<T>& v)  : Transform<T,3>(m,(Vector<T,3>)v) {};
};

/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
template<class T> class Transform4 : public Transform<T,4> {
public:
	inline Transform4() {};
	inline Transform4(const Transform<T,4>& t) : Transform<T,4>(t) {};
	inline explicit Transform4(const Matrix<T,5>& m) : Transform<T,4>(m) {};
	inline explicit Transform4(const Matrix<T,4>& m) : Transform<T,4>(m) {};
	inline explicit Transform4(const Vector<T,4>& v) : Transform<T,4>(v) {};
	inline explicit Transform4(const Vector4<T>& v)  : Transform<T,4>((Vector<T,4>)v) {};
	inline Transform4(const Matrix<T,4>& m,const Vector<T,4>& v) : Transform<T,4>(m,v) {};
	inline Transform4(const Matrix<T,4>& m,const Vector4<T>& v)  : Transform<T,4>(m,(Vector<T,4>)v) {};
};
}; };

#endif
