/* 
  Magnum 
  Copyright (C) 2002-2004 Kaya Memisoglu. All Rights Reserved.

  This software is provided 'as-is', without any express or implied
  warranty.  In no event will the authors be held liable for any damages
  arising from the use of this software.

  Permission is granted to anyone to use this software for any purpose,
  including commercial applications, and to alter it and redistribute it
  freely, under the terms in LICENSE.TXT.

  Kaya Memisoglu  (k.memisoglu@dimajix.de)
*/

#ifndef __MATH_GRAPH_H
#define __MATH_GRAPH_H

#include "container.list.h"
#include "container.array.h"


namespace magnum { namespace math {
using magnum::container::Array;
using magnum::container::List;
using magnum::object::Object;

/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
template<class N, class E> class Graph : public Object {
DeclareClass(Graph, Object);
public:
	class Node;
	class Edge;

	class Edge {
	friend Graph;
	friend Node;
	private:
		E						m_Edge;
		bool					m_Marker;
		unsigned int			m_Start;
		unsigned int			m_End;

	private:
		inline Edge(unsigned int s,unsigned int e) {
			m_Start = s;
			m_End = e;
			m_Marker = false;
		};
		inline Edge(unsigned int s,unsigned int e,const E& elem) {
			m_Start = s;
			m_End = e;
			m_Edge = elem;
			m_Marker = false;
		};

		
	public:
		inline Edge() {};
		inline unsigned int		getStart() const {
			return m_Start;
		};
		inline unsigned int		getEnd() const {
			return m_End;
		};
		
		inline E*				operator->() {
			return &m_Edge;
		};
		inline const E*			operator->() const {
			return &m_Edge;
		};
		inline E&				operator*() {
			return m_Edge;
		};
		inline const E&			operator*() const {
			return m_Edge;
		};

		inline void				mark() {
			m_Marker = true;
		};
		inline void				unmark() {
			m_Marker = false;
		};
		inline bool				isMarked() const {
			return m_Marker;
		};
	};


	class Node {
	friend Graph;
	private:
		N						m_Node;
		bool					m_Marker;
		unsigned int			m_Index;
		List<Edge>				m_Edges;

	public:
		typedef typename magnum::container::List<Edge>::Iterator Iterator;
		
	public:
		inline Node() {
			m_Marker = false;
		};

		inline N*				operator->() {
			return &m_Node;
		};
		inline const N*			operator->() const {
			return &m_Node;
		};
		inline N&				operator*() {
			return m_Node;
		};
		inline const N&			operator*() const {
			return m_Node;
		};
		
		inline Iterator			begin() const {
			return m_Edges.begin();
		};
		inline Iterator			end() const {
			return m_Edges.end();
		};
								
		inline void				mark() {
			m_Marker = true;
		};
		inline void				unmark() {
			m_Marker = false;
		};
		inline bool				isMarked() const {
			return m_Marker;
		};
								
		inline void				clear() {
			m_Edges.clear();
		};
		inline void				insert(unsigned int target) {
			m_Edges.insert(Edge(m_Index, target));
		};
		inline void				insert(unsigned int target, const E& e) {
			m_Edges.insert(Edge(m_Index, target, e));
		};
		void					remove(unsigned int target) {
			Iterator iter, end = m_Edges.end();
			for (iter = m_Edges.begin(); iter != end; iter++) {
				if (iter->m_End == target) {
					m_Edges.remove(iter);
					return;
				}
			}
		};

		inline void				remove(const Iterator& iter) {
			m_Edges.remove(iter);
		};
		bool					contains(unsigned int target) const {
			Iterator iter, end = m_Edges.end();
			for (iter = m_Edges.begin(); iter != end; iter++) {
				if (iter->m_End == target)
					return true;
			}

			return false;
		};
								
		inline bool				isEmpty() const {
			return m_Edges.isEmpty();
		};
		inline unsigned int		getSize() const {
			return m_Edges.getSize();
		};
	};
	
	
	typedef typename magnum::container::Array<Node>::Iterator Iterator;
	
private:
	Array<Node>			m_Nodes;
	
public:
	inline Graph();
	inline explicit Graph(unsigned int );

	void				setSize(unsigned int );
	inline unsigned int	getSize() const;
	
	inline Node&		operator[](unsigned int );
	inline const Node&	operator[](unsigned int ) const;
	
	inline Iterator		begin() const;
	inline Iterator		end() const;
	inline Iterator		at(unsigned int ) const;
	
	void				clear();
	inline void			insert(unsigned int start, unsigned int end);
	inline void			insert(unsigned int start, unsigned int end, const E& elem);
	inline void			insert(const Iterator& start, const Iterator& end);
	inline void			insert(const Iterator& start, const Iterator& end, const E& elem);
	inline void			remove(unsigned int start, unsigned int end);
	inline void			remove(const Iterator& start, const Iterator& end);
	inline bool			contains(unsigned int start, unsigned int end);
	inline bool			contains(const Iterator& start, const Iterator& end);
	
	bool				isEmpty() const;
};

}; };

#endif
