/* 
  Magnum 
  Copyright (C) 2002-2004 Kaya Memisoglu. All Rights Reserved.

  This software is provided 'as-is', without any express or implied
  warranty.  In no event will the authors be held liable for any damages
  arising from the use of this software.

  Permission is granted to anyone to use this software for any purpose,
  including commercial applications, and to alter it and redistribute it
  freely, under the terms in LICENSE.TXT.

  Kaya Memisoglu  (k.memisoglu@dimajix.de)
*/

#ifndef __MATH_MATRIX_H
#define __MATH_MATRIX_H

#include "math.vector.h"


namespace magnum { namespace math {

/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
template<class T, int dim> class Matrix {
protected:
	T			m_Matrix[dim][dim];

public:
	static const Matrix Identity;
	static const Matrix Null;

public:
	inline Matrix();
	inline explicit Matrix(const T* );

	inline Matrix			operator-() const;
							
	inline Matrix			operator*(const T& ) const;
	inline Matrix			operator/(const T& ) const;
							
	inline Matrix			operator+(const Matrix& ) const;
	inline Matrix			operator-(const Matrix& ) const;
	inline Vector<T,dim>	operator*(const Vector<T,dim>& ) const;
	inline Matrix			operator*(const Matrix& ) const;

	inline const Matrix&	operator*=(const T& );
	inline const Matrix&	operator/=(const T& );
	inline const Matrix&	operator+=(const Matrix& );
	inline const Matrix&	operator-=(const Matrix& );
	inline const Matrix&	operator*=(const Matrix& );

	inline bool				operator==(const Matrix& ) const;
	inline bool				operator!=(const Matrix& ) const;

	inline T*				operator[](unsigned int );
	inline const T*			operator[](unsigned int ) const;
							
	inline T&				operator()(unsigned int, unsigned int );
	inline const T&			operator()(unsigned int, unsigned int ) const;
							
	inline Matrix			getTransposed() const;
	inline Matrix			getInverse() const;
	inline Matrix&			transpose();
	inline Matrix&			invert();
	inline Matrix&			identity();
	inline Matrix&			null();

	inline Vector<T,dim>	getRow(unsigned int ) const;
	inline Vector<T,dim>	getColumn(unsigned int ) const;
	inline void				setRow(unsigned int, const Vector<T,dim>& );
	inline void				setColumn(unsigned int, const Vector<T,dim>& );
	
private:
	template<int mdim>
	inline Matrix			doGetInverse() const { return doGetInverseN(); };
	template<>
	inline Matrix			doGetInverse<2>() const { return doGetInverse2(); };
	template<int mdim>
	inline Matrix&			doInvert() { return doInvertN(); };
	template<>
	inline Matrix&			doInvert<2>() { return doInvert2(); };
	template<>
	inline Matrix&			doInvert<3>() { return doInvert3(); };
	template<>
	inline Matrix&			doInvert<4>() { return doInvert4(); };

	inline Matrix			doGetInverse2() const;
	inline Matrix			doGetInverseN() const;
	inline Matrix&			doInvert2();
	inline Matrix&			doInvert3();
	inline Matrix&			doInvert4();
	inline Matrix&			doInvertN();
};


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
template<class T, int dim> class _MatrixIdentity : public Matrix<T,dim> {
public:
	_MatrixIdentity();
};


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
template<class T, int dim> class _MatrixNull : public Matrix<T,dim> {
public:
	_MatrixNull();
};


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
template<class T> class Matrix2 : public Matrix<T, 2> {
public:
	inline Matrix2() {};
	inline Matrix2(const T* mat) : Matrix<T,2>(mat) {};
	inline Matrix2(const Matrix<T,2>& mat) : Matrix<T,2>(mat) {};

	inline const Matrix2&	operator=(const Matrix<T,2>& mtx) {Matrix<T,2>::operator=(mtx); return *this; };
};


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
template<class T> class Matrix3 : public Matrix<T, 3> {
public:
	inline Matrix3() {};
	inline Matrix3(const T* mat) : Matrix<T,3>(mat) {};
	inline Matrix3(const Matrix<T,3>& mat) : Matrix<T,3>(mat) {};

	inline const Matrix3&	operator=(const Matrix<T,3>& mtx) {Matrix<T,3>::operator=(mtx); return *this; };
};


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
template<class T> class Matrix4 : public Matrix<T, 4> {
public:
	inline Matrix4() {};
	inline Matrix4(const T* mat) : Matrix<T,4>(mat) {};
	inline Matrix4(const Matrix<T,4>& mat) : Matrix<T,4>(mat) {};

	inline const Matrix4&	operator=(const Matrix<T,4>& mtx) {Matrix<T,4>::operator=(mtx); return *this; };
};


}; };

#endif
