/* 
  Magnum 
  Copyright (C) 2002-2004 Kaya Memisoglu. All Rights Reserved.

  This software is provided 'as-is', without any express or implied
  warranty.  In no event will the authors be held liable for any damages
  arising from the use of this software.

  Permission is granted to anyone to use this software for any purpose,
  including commercial applications, and to alter it and redistribute it
  freely, under the terms in LICENSE.TXT.

  Kaya Memisoglu  (k.memisoglu@dimajix.de)
*/

#ifndef __MATH_VECTOR4_INL
#define __MATH_VECTOR4_INL

namespace magnum {namespace math {

/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
template<class T> 
Vector4<T> Vector4<T>::Null = Vector4<T>(T(0), T(0), T(0), T(0));


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
template<class T> 
inline Vector4<T>::Vector4() {
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
template<class T> 
inline Vector4<T>::Vector4(const T* s) {
	x = s[0];
	y = s[1];
	z = s[2];
	w = s[3];
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
template<class T> 
inline Vector4<T>::Vector4(const Vector<T,4>& s) {
	x = s[0];
	y = s[1];
	z = s[2];
	w = s[3];
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
template<class T> 
inline Vector4<T>::Vector4(const T& t1, const T& t2, const T& t3, const T& t4) {
	x = t1;
	y = t2;
	z = t3;
	w = t4;
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
template<class T> 
inline Vector4<T>::Vector4(const Vector3<T>& v) {
	x = v.x;
	y = v.y;
	z = v.z;
	w = T(0);
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
template<class T> 
inline Vector4<T>::Vector4(const Vector2<T>& v) {
	x = v.x;
	y = v.y;
	w = z = T(0);
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
template<class T> 
inline Vector4<T> Vector4<T>::operator+(const Vector4<T> &u) const {
	return Vector4(x + u.x, y + u.y, z + u.z, w + u.w);
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
template<class T> 
inline Vector4<T> Vector4<T>::operator+(const T& c) const {
	return Vector4(x + c, y + c, z + c, w + c);
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
template<class T> 
inline Vector4<T> Vector4<T>::operator-(const Vector4<T> &u) const {
	return Vector4(x - u.x, y - u.y, z - u.z, w - u.w);
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
template<class T> 
inline Vector4<T> Vector4<T>::operator-(const T& c) const {
	return Vector4(x - c, y - c, z - c, w - c);
}



/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
template<class T> 
inline Vector4<T> Vector4<T>::operator-() const {
	return Vector4(-x, -y, -z, -w);
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
template<class T> 
inline Vector4<T> Vector4<T>::operator*(const T& s) const {
	return Vector4(s*x, s*y, s*z, s*w);
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
template<class T> 
inline Vector4<T> operator*(const T& s,const Vector4<T>& v) {
	return Vector4<T>(s*v.x, s*v.y, s*v.z, s*v.w);
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
template<class T> 
inline Vector4<T> operator*(const Matrix4<T>& mtx, const Vector4<T>& v) {
	return Vector4<T>(
		v.x*mtx[0][0] + v.y*mtx[0][1] + v.z*mtx[0][2] + v.w*mtx[0][3], 
		v.x*mtx[1][0] + v.y*mtx[1][1] + v.z*mtx[1][2] + v.w*mtx[1][3],
		v.x*mtx[2][0] + v.y*mtx[2][1] + v.z*mtx[2][2] + v.w*mtx[2][3],
		v.x*mtx[3][0] + v.y*mtx[3][1] + v.z*mtx[3][2] + v.w*mtx[3][3]);
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
template<class T> 
inline T Vector4<T>::operator*(const Vector4<T>& v) const {
	return x*v.x + y*v.y + z*v.z + w*v.w;
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
template<class T> 
inline Vector4<T> Vector4<T>::operator/(const T& s) const {
	return Vector4(x / s, y / s, z / s, w / s);
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
template<class T> 
inline const Vector4<T>& Vector4<T>::operator+= (const T& s) {
	x += s;
	y += s;
	z += s;
	w += s;
	return *this;
}



/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
template<class T> 
inline const Vector4<T>& Vector4<T>::operator+= (const Vector4<T>& v) {
	x += v.x;
	y += v.y;
	z += v.z;
	w += v.w;
	return *this;
}



/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
template<class T> 
inline const Vector4<T>& Vector4<T>::operator-= (const T& s) {
	x -= s;
	y -= s;
	z -= s;
	w -= s;
	return *this;
}



/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
template<class T> 
inline const Vector4<T>& Vector4<T>::operator-= (const Vector4<T>& v) {
	x -= v.x;
	y -= v.y;
	z -= v.z;
	w -= v.w;
	return *this;
}



/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
template<class T> 
inline const Vector4<T>& Vector4<T>::operator*= (const T& s) {
	x *= s;
	y *= s;
	z *= s;
	w *= s;
	return *this;
}



/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
template<class T> 
inline const Vector4<T>& Vector4<T>::operator/= (const T& s) {
	x /= s;
	y /= s;
	z /= s;
	w /= s;
	return *this;
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
template<class T> 
inline const Vector4<T>& Vector4<T>::operator=(const Vector<T,4>& v) {
	x = v.vec[0];
	y = v.vec[1];
	z = v.vec[2];
	w = v.vec[3];
	return *this;
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
template<class T> 
inline const Vector4<T>& Vector4<T>::operator=(const Vector<T,3>& v) {
	x = v.vec[0];
	y = v.vec[1];
	z = v.vec[2];
	w = T(0);
	return *this;
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
template<class T> 
inline const Vector4<T>& Vector4<T>::operator=(const Vector<T,2>& v) {
	x = v.vec[0];
	y = v.vec[1];
	z = 
	w = T(0);
	return *this;
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
template<class T> 
inline const Vector4<T>& Vector4<T>::operator=(const Vector3<T>& v) {
	x = v.x;
	y = v.y;
	z = v.z;
	w = T(0);
	return *this;
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
template<class T> 
inline const Vector4<T>& Vector4<T>::operator=(const Vector2<T>& v) {
	x = v.x;
	y = v.y;
	z = 
	w = T(0);
	return *this;
};


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
template<class T> 
inline bool Vector4<T>::operator==(const Vector4<T>& v) const {
	return ((x == v.x) && (y == v.y) && (z == v.z) && (w == v.w));
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
template<class T> 
inline bool Vector4<T>::operator!=(const Vector4<T>& v) const {
	return ((x != v.x) || (y != v.y) || (z != v.z) || (w != v.w));
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
template<class T> 
inline bool Vector4<T>::operator<(const Vector4<T>& v) const {
	if (x < v.x) return true;
	if (x > v.x) return false;
	if (y < v.y) return true;
	if (y > v.y) return false;
	if (z < v.z) return true;
	if (z > v.z) return false;
	return (w < v.w);
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
template<class T> 
inline bool Vector4<T>::operator>(const Vector4<T>& v) const {
	if (x > v.x) return true;
	if (x < v.x) return false;
	if (y > v.y) return true;
	if (y < v.y) return false;
	if (z > v.z) return true;
	if (z < v.z) return false;
	return (w > v.w);
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
template<class T> 
inline bool Vector4<T>::operator<=(const Vector4<T>& v) const {
	if (x < v.x) return true;
	if (x > v.x) return false;
	if (y < v.y) return true;
	if (y > v.y) return false;
	if (z < v.z) return true;
	if (z > v.z) return false;
	return (w <= v.w);
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
template<class T> 
inline bool Vector4<T>::operator>=(const Vector4<T>& v) const {
	if (x > v.x) return true;
	if (x < v.x) return false;
	if (y > v.y) return true;
	if (y < v.y) return false;
	if (z > v.z) return true;
	if (z < v.z) return false;
	return (w >= v.w);
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
template<class T> 
inline Vector2<T>	Vector4<T>::xy() const {
	return Vector2<T>(x,y);
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
template<class T> 
inline Vector2<T>	Vector4<T>::xz() const {
	return Vector2<T>(x,z);
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
template<class T> 
inline Vector2<T>	Vector4<T>::yz() const {
	return Vector2<T>(y,z);
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
template<class T> 
inline Vector2<T>	Vector4<T>::xw() const {
	return Vector2<T>(x,w);
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
template<class T> 
inline Vector2<T>	Vector4<T>::yw() const {
	return Vector2<T>(y,w);
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
template<class T> 
inline Vector2<T>	Vector4<T>::zw() const {
	return Vector2<T>(z,w);
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
template<class T> 
inline Vector3<T>	Vector4<T>::xyz() const {
	return Vector3<T>(x,y,z);
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
template<class T> 
inline Vector3<T>	Vector4<T>::xyw() const {
	return Vector3<T>(x,y,w);
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
template<class T> 
inline Vector3<T>	Vector4<T>::yzw() const {
	return Vector3<T>(y,z,w);
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
template<class T> 
inline Vector4<T>::operator const T*() const {
	return &x;
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
template<class T> 
inline Vector4<T>::operator T*() {
	return &x;
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
template<class T> 
inline Vector4<T>::operator Vector<T,4>() const {
	return Vector<T,4>(&x);
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
template<class T> 
inline Vector4<T>& Vector4<T>::null() {
	x = y = z = w = T(0);
	return *this;
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
template<class T> 
inline Vector4<T>& Vector4<T>::normalize() {
	T l = x*x + y*y + z*z + w*w;
	if (l > T(0)) {
		l = T(1)/sqrt(l);
		x *= l;
		y *= l;
		z *= l;
		w *= l;
	}
	return *this;
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
template<class T> 
inline T Vector4<T>::length() const {
	return (T)sqrt(x*x + y*y + z*z + w*w);
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
template<class T> 
inline T Vector4<T>::length2() const {
	return (x*x + y*y + z*z + w*w);
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
template<class T> 
inline T Vector4<T>::max() const {
	T a = (x > y) ? x : y;
	T b = (z > w) ? z : w;
	return (a > b) ? a : b;
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
template<class T> 
inline T Vector4<T>::min() const {
	T a = (x < y) ? x : y;
	T b = (z < w) ? z : w;
	return (a < b) ? a : b;
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
template<class T> 
inline T Vector4<T>::maxabs() const {
	T xa = abs(x);
	T ya = abs(y);
	T za = abs(z);
	T wa = abs(w);
	T a = (xa > ya) ? xa : ya;
	T b = (za > wa) ? za : wa;
	return (a > b) ? a : b;
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
template<class T> 
inline T Vector4<T>::minabs() const {
	T xa = abs(x);
	T ya = abs(y);
	T za = abs(z);
	T wa = abs(w);
	T a = (xa < ya) ? xa : ya;
	T b = (za < wa) ? za : wa;
	return (a < b) ? a : b;
}


}; };

#endif
