/* 
  Magnum 
  Copyright (C) 2002-2004 Kaya Memisoglu. All Rights Reserved.

  This software is provided 'as-is', without any express or implied
  warranty.  In no event will the authors be held liable for any damages
  arising from the use of this software.

  Permission is granted to anyone to use this software for any purpose,
  including commercial applications, and to alter it and redistribute it
  freely, under the terms in LICENSE.TXT.

  Kaya Memisoglu  (k.memisoglu@dimajix.de)
*/

#ifndef __SCENE3D_TRAVERSER_INL
#define __SCENE3D_TRAVERSER_INL

namespace magnum { namespace scene3d {


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
template<class T, class L> 
Traverser<T,L>::Traverser() {
	m_Universe = NULL;
	m_Locale = NULL;
	m_PreTransform.identity();
	m_PreTranslation.null();
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
template<class T, class L> 
Traverser<T,L>::Traverser(const Locale* l) {
	ASSERT(l != NULL);
	m_Locale = l
	m_Universe = l->getUniverse();
	m_PreTransform.identity();
	m_PreTranslation.null();
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
template<class T, class L> 
Traverser<T,L>::Traverser(const Universe* u) {
	ASSERT(u != NULL);
	m_Universe = u;
	m_Locale = NULL;
	m_PreTransform.identity();
	m_PreTranslation.null();
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
template<class T, class L> 
inline const Locale* Traverser<T,L>::getLocale() const {
	return m_Locale;
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
template<class T, class L> 
inline const Universe* Traverser<T,L>::getUniverse() const {
	return m_Universe;
}



/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
template<class T, class L> 
inline void Traverser<T,L>::setPreTransform(const Transform3<float>& t) {
	m_PreTransform = t;
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
template<class T, class L> 
inline const Transform3<float>& Traverser<T,L>::getPreTransform() const {
	return m_PreTransform;
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
template<class T, class L> 
inline void Traverser<T,L>::setPreTranslation(const Vector3<double>& t) {
	m_PreTranslation = t;
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
template<class T, class L> 
inline const Vector3<double>& Traverser<T,L>::getPreTranslation() const {
	return m_PreTranslation;
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
template<class T, class L> 
inline const Transform3<float>& Traverser<T,L>::getTransform() const {
	return m_Transform;
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
template<class T, class L> 
inline const Vector3<double>& Traverser<T,L>::getTranslation() const {
	return m_Translation;
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
template<class T, class L>
template<class V> 
void Traverser<T,L>::traverse(const V& vol, const Vector3<double>& volorg) {
	V volume(vol);
	Vector3<float> origin = volume.getOrigin();
	const Space<float, 3, L>* space;

	// Check if we only need to traverse a single locale
	if (m_Locale != NULL) {
		// Take pretransformations into account
		m_Translation = m_PreTranslation + m_Locale->getOrigin();
		m_Transform = m_PreTransform;
		m_Transform.translate((Vector3<float>)m_Translation);

		// Translate volume
		volume.setOrigin(origin + (Vector3<float>)(volorg - m_Locale->getOrigin()));

		// Intersect space!
		m_Locale->getSpace(space);
		if(volume.intersect(space->getBounds()))
			space->intersect(volume, *this);
		return;
	}

	if (m_Universe == NULL)
		return;

	// Traverse complete universe
	Universe::Iterator iter, end;
	end = m_Universe->end();
	for (iter = m_Universe->begin(); iter != end; iter++) {
		m_Locale = (Locale*)iter;

		// Take pretransformations into account
		m_Translation = m_PreTranslation + m_Locale->getOrigin();
		m_Transform = m_PreTransform;
		m_Transform.translate((Vector3<float>)m_Translation);

		// Translate volume
		volume.setOrigin(origin + (Vector3<float>)(volorg - m_Locale->getOrigin()));

		// Intersect space!
		m_Locale->getSpace(space);
		if(volume.intersect(space->getBounds()))
			space->intersect(volume, *this);
	}

	m_Locale = NULL;
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
template<class T, class L>
void Traverser<T,L>::traverse(const Vector3<float>& position, const Vector3<double>& porg) {
	const Space<float, 3, L>* space;
	Vector<float,3> p;

	// Check if we only need to traverse a single locale
	if (m_Locale != NULL) {
		// Take pretransformations into account
		m_Translation = m_PreTranslation + m_Locale->getOrigin();
		m_Transform = m_PreTransform;
		m_Transform.translate((Vector3<float>)m_Translation);

		// Translate intersection point
		p = position + (Vector3<float>)(porg - m_Locale->getOrigin());

		// Intersect space!
		m_Locale->getSpace(space);
		if (space->getBounds().isInside(p))
			space->intersect(p, *this);
		return;
	}

	if (m_Universe == NULL)
		return;

	// Traverse complete universe
	Universe::Iterator iter, end;
	end = m_Universe->end();
	for (iter = m_Universe->begin(); iter != end; iter++) {
		m_Locale = (Locale*)iter;

		// Take pretransformations into account
		m_Translation = m_PreTranslation + m_Locale->getOrigin();
		m_Transform = m_PreTransform;
		m_Transform.translate((Vector3<float>)m_Translation);

		// Translate intersection point
		p = position + (Vector3<float>)(porg - m_Locale->getOrigin());

		// Intersect space!
		m_Locale->getSpace(space);
		if (space->getBounds().isInside(p))
			space->intersect(p, *this);
	}

	m_Locale = NULL;
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
template<class T, class L> 
inline void Traverser<T,L>::onVisit(L* ) {
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
template<class T, class L> 
inline void Traverser<T,L>::operator()(L* item) {
	((T*)this)->onVisit(item);
}


}; };

#endif
