/* 
  Magnum 
  Copyright (C) 2002-2004 Kaya Memisoglu. All Rights Reserved.

  This software is provided 'as-is', without any express or implied
  warranty.  In no event will the authors be held liable for any damages
  arising from the use of this software.

  Permission is granted to anyone to use this software for any purpose,
  including commercial applications, and to alter it and redistribute it
  freely, under the terms in LICENSE.TXT.

  Kaya Memisoglu  (k.memisoglu@dimajix.de)
*/

#include "io.exception.h"
#include "graphics.translator.gif.h"

#include "object.inl"
#include "object.exception.inl"
#include "base.string.inl"
#include "graphics.translator.inl"
#include "graphics.colour.inl"
#include "graphics.palette.inl"
#include "graphics.image.inl"
#include "graphics.dimension.inl"

#if defined(__FORWINDOWS__)
#undef DrawText
#endif

#include "gif_lib.h"

namespace magnum { namespace graphics {

static unsigned int InterlacedOffset[] = { 0, 4, 2, 1 }; /* The way Interlaced image should. */
static unsigned int InterlacedJumps[] = { 8, 8, 4, 2 };    /* be read - offsets and jumps... */

#define MCHAR2(a, b) (a | (b << 8))
#define MCHAR4(a, b, c, d) (a | (b << 8) | (c << 16) | (d << 24))

/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
static int GifReadFunc(GifFileType* gf, GifByteType* buffer, int size) {
	InputStream* in = (InputStream*)gf->UserData;
	return in->read(buffer, size);
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
Image* TranslatorGIF::onDecode(InputStream* in, long size) {
	GifFileType* file;
	Image* img;
	byte* imgbuf;
	unsigned int i, j, lsize, Row, Col, Width, Height, Count;
	int ExtCode;

	file = DGifOpen(in, GifReadFunc);
	if (file == NULL)
		throw magnum::io::TranslatorFormatException();

	
    /* Allocate the screen as vector of column of rows. We cannt allocate    */
    /* the all screen at once, as this broken minded CPU can allocate up to  */
    /* 64k at a time and our image can be bigger than that:		     */
    /* Note this screen is device independent - its the screen as defined by */
    /* the GIF file parameters itself.					     */
	GifRowType *ScreenBuffer = new GifRowType[file->SHeight];
    if (ScreenBuffer == NULL)
		throw magnum::io::TranslatorMemoryException();

    lsize = file->SWidth * sizeof(GifPixelType);

	img = new Image(Image::Index8, file->SWidth, file->SHeight);
	imgbuf = (byte*)img->lock();

	for (i = 0; i < (unsigned int)file->SHeight; i++) {
		byte* row = imgbuf + i*img->getLinesize();
		ScreenBuffer[i] = (GifRowType)row;
		memset(ScreenBuffer[i], file->SBackGroundColor, lsize);
	}

    /* Scan the content of the GIF file and load the image(s) in: */
	GifByteType *Extension;
	GifRecordType RecordType;

    do {
		if (DGifGetRecordType(file, &RecordType) == GIF_ERROR)
			throw magnum::io::TranslatorFormatException();
		switch (RecordType) {
			case IMAGE_DESC_RECORD_TYPE:
				if (DGifGetImageDesc(file) == GIF_ERROR)
					throw magnum::io::TranslatorFormatException();
				Row = file->Image.Top; /* Image Position relative to Screen. */
				Col = file->Image.Left;
				Width = file->Image.Width;
				Height = file->Image.Height;
				if (file->Image.Left + file->Image.Width > file->SWidth ||
				   file->Image.Top + file->Image.Height > file->SHeight) {
					throw magnum::io::TranslatorFormatException();
				}
				if (file->Image.Interlace) {
					/* Need to perform 4 passes on the images: */
					for (Count = i = 0; i < 4; i++)
						for (j = Row + InterlacedOffset[i]; j < Row + Height; j += InterlacedJumps[i]) {
							if (DGifGetLine(file, &ScreenBuffer[j][Col], Width) == GIF_ERROR) {
								throw magnum::io::TranslatorFormatException();
							}
						}
				} else {
					for (i = 0; i < Height; i++) {
						if (DGifGetLine(file, &ScreenBuffer[Row++][Col], Width) == GIF_ERROR)
							throw magnum::io::TranslatorFormatException();
					}
				}
				break;
			case EXTENSION_RECORD_TYPE:
				/* Skip any extension blocks in file: */
				if (DGifGetExtension(file, &ExtCode, &Extension) == GIF_ERROR)
					throw magnum::io::TranslatorFormatException();
				while (Extension != NULL) {
					if (DGifGetExtensionNext(file, &Extension) == GIF_ERROR)
						throw magnum::io::TranslatorFormatException();
				}
				break;
			case TERMINATE_RECORD_TYPE:
				break;
			default:		    /* Should be traps by DGifGetRecordType. */
				break;
		}
    } while (RecordType != TERMINATE_RECORD_TYPE);

	img->unlock();

	GifColorType *ColorMap = (file->Image.ColorMap ?
                    file->Image.ColorMap->Colors :
                    file->SColorMap->Colors);
    int ColorMapSize = 1 << (file->Image.ColorMap ? file->Image.ColorMap->BitsPerPixel :
		            file->SColorMap->BitsPerPixel);

	Palette<byte> pal;
	for (i = 0; i < (unsigned int)ColorMapSize; i++) 
		pal.setColour(Colour<byte>(ColorMap[i].Red, ColorMap[i].Green, ColorMap[i].Blue), i);
	img->setPalette(pal);

    if (DGifCloseFile(file) == GIF_ERROR)
		throw magnum::io::TranslatorFormatException();
	
	return img;
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
long TranslatorGIF::onEncode(OutputStream* out, const Image* img) {
	return 0;
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
bool TranslatorGIF::onIdentify(File* in) {
	dword id;

	try {
		unsigned long oldpos;
		oldpos = in->getPosition();	
		in->read(&id, sizeof(id));
		in->setPosition(oldpos, File::Absolute);
	} catch (...) {
		return false;
	}

	if (id != MCHAR4('G', 'I', 'F', '8'))
		return false;

	return true;
} 


}; };
