/* 
  Magnum 
  Copyright (C) 2002-2004 Kaya Memisoglu. All Rights Reserved.

  This software is provided 'as-is', without any express or implied
  warranty.  In no event will the authors be held liable for any damages
  arising from the use of this software.

  Permission is granted to anyone to use this software for any purpose,
  including commercial applications, and to alter it and redistribute it
  freely, under the terms in LICENSE.TXT.

  Kaya Memisoglu  (k.memisoglu@dimajix.de)
*/

#include "io.exception.h"
#include "graphics.translator.png.h"

#include "object.inl"
#include "object.exception.inl"
#include "base.string.inl"
#include "io.translator.inl"
#include "container.chain.inl"
#include "graphics.translator.inl"
#include "graphics.palette.inl"
#include "graphics.colour.inl"
#include "graphics.image.inl"
#include "graphics.dimension.inl"

#include "png.h"

#include <setjmp.h>


namespace magnum { namespace graphics {

/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
static void png_read_data(png_structp m_Png, png_bytep data, png_size_t length) {
	InputStream* in = (InputStream*)png_get_io_ptr(m_Png);
	in->read(data, (unsigned long)length);
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
TranslatorPNG::TranslatorPNG() {
    m_Png = png_create_read_struct(PNG_LIBPNG_VER_STRING, NULL, NULL, (png_error_ptr)NULL);
    m_Info = png_create_info_struct(m_Png);
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
TranslatorPNG::~TranslatorPNG() {
	png_destroy_read_struct(&m_Png, &m_Info, NULL);
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
Image* TranslatorPNG::onDecode(InputStream* in, long size) {
	Image* img = NULL;
    png_byte pbSig[8];
    int iBitDepth;
	png_uint_32 iHeight, iWidth;
    int iColorType;
    double dGamma;
    png_color_16 *pBackground;
    png_uint_32 ulChannels;
    png_uint_32 ulRowBytes;
    png_byte *pbImageData;
	png_byte **ppbRowPointers = NULL;
    png_uint_32 i;

    // first check the eight byte PNG signature
    in->read(pbSig, 8);
    if (!png_check_sig(pbSig, 8))
		throw magnum::io::TranslatorFormatException();

	// Install Error handler
    if (setjmp(png_jmpbuf(m_Png)))
		throw magnum::io::TranslatorFormatException();

    // initialize the png structure
    png_set_read_fn(m_Png, (png_voidp)in, png_read_data);
    png_set_sig_bytes(m_Png, 8);
    
    // read all PNG info up to image data
    png_read_info(m_Png, m_Info);
    
    // get width, height, bit-depth and color-type
    png_get_IHDR(m_Png, m_Info, &iWidth, &iHeight, &iBitDepth,
        &iColorType, NULL, NULL, NULL);
    
    // expand images of all color-type and bit-depth to 3x8 bit RGB images
    // let the library process things like alpha, transparency, background
    if (iBitDepth == 16)
        png_set_strip_16(m_Png);
    if (iBitDepth < 8)
        png_set_packing(m_Png);
    if (png_get_valid(m_Png, m_Info, PNG_INFO_tRNS))
        png_set_expand(m_Png);
    
    // set the background color to draw transparent and alpha images over.
    if (png_get_bKGD(m_Png, m_Info, &pBackground)) {
        png_set_background(m_Png, pBackground, PNG_BACKGROUND_GAMMA_FILE, 1, 1.0);
//        pBkgColor->red   = (byte) pBackground->red;
//        pBkgColor->green = (byte) pBackground->green;
//        pBkgColor->blue  = (byte) pBackground->blue;
    } else {
//        pBkgColor = NULL;
    }
    
    // if required set gamma conversion
    if (png_get_gAMA(m_Png, m_Info, &dGamma))
        png_set_gamma(m_Png, (double) 2.2, dGamma);
    
    // after the transformations have been registered update m_Info data
    png_read_update_info(m_Png, m_Info);
    
    // get again width, height and the new bit-depth and color-type
    png_get_IHDR(m_Png, m_Info, &iWidth, &iHeight, &iBitDepth,
        &iColorType, NULL, NULL, NULL);

    ulChannels = png_get_channels(m_Png, m_Info);

    // now we can allocate memory to store the image
	Image::Type type;
	switch (iColorType) {
		case PNG_COLOR_TYPE_PALETTE:
			type = Image::Index8;
			break;
	    case PNG_COLOR_TYPE_GRAY:
			type = Image::GR8;
			break;
		case PNG_COLOR_TYPE_GRAY_ALPHA:
			type = Image::GRA8;
			break;
		case PNG_COLOR_TYPE_RGB:
			if (ulChannels == 4)
				type = Image::RGBA8;
			else
				type = Image::RGB8;
			break;
		case PNG_COLOR_TYPE_RGB_ALPHA:
			type = Image::RGBA8;
			break;
		default:
			throw magnum::io::TranslatorFormatException();
	}

	img = new Image(type, iWidth, iHeight);
	if (img == NULL)
		throw magnum::io::TranslatorMemoryException();

	// get any palette
	int paletteSize = 0;
	png_color* palette = NULL;
	if (png_get_PLTE(m_Png, m_Info, &palette, &paletteSize)) {
		if (paletteSize > 0) {
			Palette<byte> pal;
			for (i = 0; i < (unsigned int)paletteSize; i++)
				pal[i] = Colour<byte>(palette[i].red, palette[i].green, palette[i].blue);
			img->setPalette(pal);
		}
	}


    // and allocate memory for an array of row-pointers
    ppbRowPointers = (png_bytepp) new png_bytep[iHeight];
	if (ppbRowPointers == NULL)
		throw magnum::io::TranslatorMemoryException();

	pbImageData = (png_byte *)img->lock();
	ulRowBytes = img->getLinesize();
    
    // set the individual row-pointers to point at the correct offsets
    for (i = 0; i < iHeight; i++)
        ppbRowPointers[i] = pbImageData + i * ulRowBytes;
    
    // now we can go ahead and just read the whole image
    png_read_image(m_Png, ppbRowPointers);
    
    // read the additional chunks in the PNG file (not really needed)
    //png_read_end(m_Png, NULL);
    
    // and we're done
	img->unlock();
    delete[] ppbRowPointers;

	return img;
}



/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
long TranslatorPNG::onEncode(OutputStream* out, const Image* img) {
	return 0;
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
bool TranslatorPNG::onIdentify(File* in) {
    png_byte pbSig[8];

	try {
		unsigned long oldpos;
		oldpos = in->getPosition();	
		in->read(&pbSig, 8);
		in->setPosition(oldpos, File::Absolute);
	} catch (...) {
		return false;
	}

    if (!png_check_sig(pbSig, 8))
		return false;

	return true;
} 


}; };
