/* 
  Magnum 
  Copyright (C) 2002-2004 Kaya Memisoglu. All Rights Reserved.

  This software is provided 'as-is', without any express or implied
  warranty.  In no event will the authors be held liable for any damages
  arising from the use of this software.

  Permission is granted to anyone to use this software for any purpose,
  including commercial applications, and to alter it and redistribute it
  freely, under the terms in LICENSE.TXT.

  Kaya Memisoglu  (k.memisoglu@dimajix.de)
*/

#include "io.exception.h"
#include "graphics.translator.tga.h"

#include "object.inl"
#include "object.exception.inl"
#include "base.string.inl"
#include "graphics.image.inl"
#include "graphics.translator.inl"
#include "graphics.colour.inl"
#include "graphics.palette.inl"
#include "graphics.dimension.inl"


namespace magnum { namespace graphics {

#pragma pack(1)
struct TargaHeader {
	byte	ID;
	byte	CMap;
	byte	Type;
	word	FirstCol;
	word	ColLen;
	byte	ColSize;
	word	XOrig;
	word	YOrig;
	word	XSize;
	word	YSize;
	byte	BitsPerPixel;
	byte	ImageBits;
};


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
Image* TranslatorTGA::onDecode(InputStream* in, long size) {
	TargaHeader header;
	Image* img;
	int width,height,depth,bpp,i,k;
	word TempWord;
	dword TempLong;
	byte TempBytes[8];
	byte* data;

	if (size == Autosize)
		size = in->available();

	in->read(&header, sizeof(header));
	size -= sizeof(header);

	width = header.XSize;
	height = header.YSize;
	depth = header.BitsPerPixel;
	if (depth < 8)
		depth = 8;
	bpp = (depth+7)/8;
	if (bpp > 32)
		throw magnum::io::TranslatorFormatException();

	if ((header.CMap == 0) && (depth <= 8)) {
		img = new Image(Image::Index8, width, height);
	} else if (depth == 15) {
		img = new Image(Image::RGB5, width, height);
	} else if (depth == 16) {
		img = new Image(Image::R5G6B5, width, height);
	} else if (depth == 24) {
		img = new Image(Image::BGR8, width, height);
	} else if (depth == 32) {
		img = new Image(Image::BGRA8, width, height);
	} else {
		throw magnum::io::TranslatorFormatException();
	}

	if (img == NULL)
		throw magnum::io::TranslatorMemoryException();

	// Eine eventuelle Beschreibung wird geskippt...
	in->skip(header.ID);
	size -= header.ID;

	// Wir haben alle Daten, jetzt lesen wir das Image...
	if (header.CMap == 1){
		Palette<byte> pal;
		switch (header.ColSize){
			case 15 :
			case 16 :
				for (i=0; i<header.ColLen; i++) {
					in->read(&TempWord, sizeof(TempWord));
					pal[i] = Colour<byte>(
						(TempWord & 0x7c00) >> 7,
						(TempWord & 0x3e0) >> 2,
						(TempWord & 0x1f) << 3);
				}
				size -= header.ColLen*2;
				break;
			case 24 :
				for (i=0; i<header.ColLen; i++) {
					in->read(TempBytes, 3);
					pal[i] = Colour<byte>(TempBytes[2], TempBytes[1], TempBytes[0]);
				}
				size -= header.ColLen*3;
				break;
			case 32 :
				for (i=0; i<header.ColLen; i++) {
					in->read(TempBytes, 4);
					pal[i] = Colour<byte>(TempBytes[2], TempBytes[1], TempBytes[0], TempBytes[3]);
				}
				size -= header.ColLen*4;
				break;
			default:
				img->release();
				throw magnum::io::TranslatorFormatException();
		}

		img->setPalette(pal);
	}

	data = (byte *)img->lock();
	switch (header.Type){
		case 0x02 :
		case 0x01 :
			in->read(data, width*height*bpp);
			break;
		case 0x0a :
			k = 0;
/*			while (k < bpp*width*height){
				in->read(&cnt, 1);
				if (cnt & 0x80){
					cnt &= ~0x80;
					// Wir mssen einen Pixel einlesen...
					in->read(TempBytes, bpp);
					for (; cnt>=0; cnt--) {
						memcpy(data + k,TempBytes, bpp);
						k+=bpp;
					}
				} else {
					file.read (Picture->Data+k,bpp*(i+1));
					k+=(i+1)*bpp;
				}
			}*/
			break;
		default:
			img->unlock();
			img->release();
			throw magnum::io::TranslatorFormatException();
	}
	img->unlock();


	// Wenn das Bild ein 15Bit Targa war, dann sind R und B vertauscht.
	// Also mssen wir die wieder zurcktauschen...
	if ((img->getType() == Image::RGB5) || (img->getType() == Image::R5G6B5)){
		word* data = (word *)img->lock();
		for (i = 0; i < width*height; i++){
			TempLong = *data;
			TempLong = ((TempLong & 0x1f)<<10) | ((TempLong & 0x7c00)>>10) | (TempLong & 0x03e0);
			*data++ = (unsigned short)TempLong;
		}
		img->unlock();
	}

	// Das Bild ist jetzt eingeladen und entpackt und so weiter.
	// Es kann allerdings sein, da es in X- und/oder Y-Richtung
	// geflippt werden mu...
	if (header.ImageBits & 0x10){
		// X-Flippen
	}
	if (!(header.ImageBits & 0x20)){
		// Y-Flippen
		flipY(img);
	}

	return img;
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
long TranslatorTGA::onEncode(OutputStream* out, const Image* img) {
	return 0;
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
bool TranslatorTGA::onIdentify(File* in) {
	dword id;

	try {
		unsigned long oldpos;
		oldpos = in->getPosition();	
		in->read(&id, sizeof(id));
		in->setPosition(oldpos, File::Absolute);
	} catch (...) {
		return false;
	}

	if ((((unsigned char)(id >> 8) == 0) || ((unsigned char)(id >> 8) == 1)) 
		&& ((unsigned char )(id >> 16) <= 11))
		return true;

	return false;
} 


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
void TranslatorTGA::flipY(Image* img){
	byte* buffer;
	byte* data;
	int	y;
	int ls = img->getLinesize();
	int bpp = (img->getPixelsize() + 7) / 8;
	int h = img->getHeight();

	buffer = new byte[img->getLinesize()];
	data = (byte *)img->lock();
	for (y = 0; y < h/2; y++){
		memcpy(buffer, data + y*ls, ls);
		memcpy(data + y*ls,data + (h-1-y)*ls, ls);
		memcpy(data + (h-1-y)*ls, buffer, ls);
	}
	img->unlock();
	delete[] buffer;
}

}; };
