/* 
  Magnum 
  Copyright (C) 2002-2004 Kaya Memisoglu. All Rights Reserved.

  This software is provided 'as-is', without any express or implied
  warranty.  In no event will the authors be held liable for any damages
  arising from the use of this software.

  Permission is granted to anyone to use this software for any purpose,
  including commercial applications, and to alter it and redistribute it
  freely, under the terms in LICENSE.TXT.

  Kaya Memisoglu  (k.memisoglu@dimajix.de)
*/

#include "graphics3d.meshbuilder.h"
#include "graphics3d.indexmesh.h"
#include "graphics3d.indexbuffer.h"
#include "graphics3d.vertexbuffer.h"
#include "graphics3d.context.h"
#include "graphics3d.appearance.h"

#include "object.inl"
#include "object.pointer.inl"
#include "graphics3d.resource.inl"
#include "graphics3d.meshbuilder.inl"
#include "graphics3d.indexmesh.inl"
#include "graphics3d.indexbuffer.inl"
#include "graphics3d.vertexbuffer.inl"


namespace magnum { namespace graphics3d {

/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
MeshBuilder::MeshBuilder() {
	clear();
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
void MeshBuilder::setAttributes(unsigned int a) {
	m_Attributes = a & (Attributes::Vertex | Attributes::Normal | Attributes::Colour | Attributes::Texture);
	m_VertexFormat.clear();
	if (m_Attributes & Attributes::Vertex)
		m_VertexFormat.insert(Types::Float3, Attributes::Vertex);
	if (m_Attributes & Attributes::Normal)
		m_VertexFormat.insert(Types::Float3, Attributes::Normal);
	if (m_Attributes & Attributes::Colour)
		m_VertexFormat.insert(Types::Float3, Attributes::Colour);
	if (m_Attributes & Attributes::Texture)
		m_VertexFormat.insert(Types::Float2, Attributes::Texture);
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
void MeshBuilder::clear() {
	m_Primitive = Mesh::TriangleList;
	m_MaxIndex = 0;
	m_Current.m_Vertex.null();
	m_Current.m_Normal.x = 1.0f;
	m_Current.m_Normal.y = 0.0f;
	m_Current.m_Normal.z = 0.0f;
	m_Current.m_TexCoord.null();
	m_Current.m_Colour.R = 1.0f;
	m_Current.m_Colour.G = 1.0f;
	m_Current.m_Colour.B = 1.0f;
	m_Indices.clear();
	m_Vertices.clear();
	setAttributes(Attributes::Vertex);
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
void MeshBuilder::finish() {
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
Mesh* MeshBuilder::build() {
	unsigned int i;

	if (m_Vertices.getSize() == 0)
		return NULL;
		
	// Create Vertex Buffer
	VertexBuffer* vb;
	vb = VertexBuffer::create(m_VertexFormat, m_Vertices.getSize(), VertexBuffer::Static);
	if (vb == NULL)
		return NULL;
		
	// Copy vertex buffer
	void* target = vb->lock(VertexBuffer::Discard);
	for (i = 0; i < m_Vertices.getSize(); i++) {
		if (m_Attributes & Attributes::Vertex) {
			*(Vector3<float>*)target = m_Vertices[i].m_Vertex;
			target = (byte*)target + sizeof(Vector3<float>);
		}
		if (m_Attributes & Attributes::Normal) {
			*(Vector3<float>*)target = m_Vertices[i].m_Normal;
			target = (byte*)target + sizeof(Vector3<float>);
		}
		if (m_Attributes & Attributes::Colour) {
			((float*)target)[0] = m_Vertices[i].m_Colour.R;
			((float*)target)[1] = m_Vertices[i].m_Colour.G;
			((float*)target)[2] = m_Vertices[i].m_Colour.B;
			target = (byte*)target + 3*sizeof(float);
		}
		if (m_Attributes & Attributes::Texture) {
			*(Vector2<float>*)target = m_Vertices[i].m_TexCoord;
			target = (byte*)target + sizeof(Vector2<float>);
		}
	}
	vb->unlock();

	// Create index buffer	
	IndexBuffer* ib = NULL;
	if (m_Indices.getSize() > 0) {
		ib = IndexBuffer::create(IndexBuffer::Index32, m_Indices.getSize(), IndexBuffer::Static);
		ib->setBuffer(0, m_Indices.getArray(), m_Indices.getSize());
	}
	
	// Create Mesh
	VertexMesh* mesh;
	if (ib != NULL) {
		mesh = new IndexMesh();
		((IndexMesh*)mesh)->attachIndexBuffer(ib);
		ib->release();
	} else {
		mesh = new VertexMesh();
	}
	
	// attach vertex buffer
	mesh->attachVertexBuffer(vb);
	vb->release();

	// Set primitive
	mesh->setPrimitive(m_Primitive);
	
	// Attach appearance
	if (m_Appearance != NULL)
		mesh->attachAppearance(m_Appearance);
	
	return mesh;
}

}; };
