/* 
  Magnum 
  Copyright (C) 2002-2004 Kaya Memisoglu. All Rights Reserved.

  This software is provided 'as-is', without any express or implied
  warranty.  In no event will the authors be held liable for any damages
  arising from the use of this software.

  Permission is granted to anyone to use this software for any purpose,
  including commercial applications, and to alter it and redistribute it
  freely, under the terms in LICENSE.TXT.

  Kaya Memisoglu  (k.memisoglu@dimajix.de)
*/

#include "object.autopointer.h"
#include "graphics3d.context.h"
#include "graphics3d.vertexprogram.h"
#include "graphics3d.pixelprogram.h"
#include "graphics3d.appearance.h"
#include "graphics3d.material.h"
#include "graphics3d.shader.h"
#include "graphics3d.springmodel.h"
#include "graphics3d.indexbuffer.h"
#include "graphics3d.vertexbuffer.h"
#include "physics.spring.h"
#include "physics.weight.h"

#include "object.pointer.inl"
#include "object.autopointer.inl"
#include "container.iterator.inl"
#include "container.allocator.inl"
#include "container.array.inl"
#include "container.list.inl"
#include "graphics3d.appearance.inl"
#include "graphics3d.context.inl"
#include "graphics3d.resource.inl"
#include "graphics3d.springmodel.inl"
#include "graphics3d.dynamicmodel.inl"
#include "graphics3d.indexbuffer.inl"
#include "graphics3d.vertexbuffer.inl"
#include "graphics3d.mesh.inl"
#include "physics.spring.inl"
#include "physics.weight.inl"
#include "physics.relaxator.inl"

namespace magnum { namespace graphics3d {
using magnum::physics::Weight;
using magnum::object::AutoPointer;
using magnum::geometry::AlignedBox3;

/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
SpringModel* SpringModel::createRectangularModel(unsigned int w, unsigned int h) {
	AutoPointer<SpringModel> model = new SpringModel();
	if (model == NULL)
		return NULL;
	AutoPointer<Relaxator<float,3> > springs = new Relaxator<float,3>();
	if (springs == NULL)
		return NULL;
	model->attachSprings(springs);

	// Create all springs
	unsigned int i,j;
	springs->setSize(w*h);
	
	// First row
	springs->connect(0, 1);
	springs->connect(0, w);
	springs->connect(0, w+1);
	for (j = 1; j < w-1; j++) {
		springs->connect(j, j+1);
		springs->connect(j, w+j);
		springs->connect(j, w+j+1);
		springs->connect(j, w+j-1);
	}
	springs->connect(w-1, w+w-1);
	springs->connect(w-1, w+w-2);
	
	// Rows inbetween
	for (i = 1; i < h-1; i++) {
		springs->connect(i*w, i*w+1);
		springs->connect(i*w, (i+1)*w);
		springs->connect(i*w, (i+1)*w+1);
		for (j = 1; j < w-1; j++) {
			springs->connect(i*w+j, i*w+(j+1));
			springs->connect(i*w+j, (i+1)*w+j);
			springs->connect(i*w+j, (i+1)*w+j+1);
			springs->connect(i*w+j, (i+1)*w+j-1);
		}
		springs->connect((i+1)*w-1, (i+2)*w-1);
		springs->connect((i+1)*w-1, (i+2)*w-2);
	}
	
	// Last row
	for (j = 0; j < w-1; j++) {
		springs->connect((h-1)*w+j, (h-1)*w+j+1);
	}
	
	// Set positions of all weights
	for (i = 0; i < h; i++) {
		for (j = 0; j < w; j++) {
			Weight<float,3>* weight = &(*springs)[w*i+j];
			weight->setPosition(10.0f*Vector3<float>((float)i - (float)h/2.0f, (float)j - (float)w/2.0f, 0.0f));
		}
	}
	
	// Reset forces and natural lengths of all springs
	springs->reset();
	
	// Create geometry
	Array<unsigned int> geom;
	geom.setSize((w-1)*(h-1)*6);
	unsigned int* idx = geom.getArray();
	for (i = 0; i < h-1; ++i) {
		for (j = 0; j < w-1; ++j) {
			*idx++ = i*w + j;
			*idx++ = (i+1)*w + j;
			*idx++ = i*w + j+1;
			
			*idx++ = i*w + j+1;
			*idx++ = (i+1)*w + j;
			*idx++ = (i+1)*w + j+1;
		}
	}
	model->setGeometry(geom.getArray(), geom.getSize());

	model->refer();
	return model;
}



/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
SpringModel::SpringModel() {
	ParameterFormat fmt;
	fmt.insert(Parameters::Position, Types::Matrix4f, 0);
	clearParameters();
	publishParameters(fmt);
	
	clearJoints();

	Bounds3<float> bounds;
	bounds.empty();
	publishBounds(bounds);

	m_Material.setAmbient(Colour<float>(0.2f, 0.1f, 0.1f));
	m_Material.setDiffuse(Colour<float>(1.0f, 0.2f, 0.1f));
	m_Material.setSpecular(Colour<float>(1.0f, 0.8f, 0.7f));
	m_Material.setEmission(Colour<float>(0.0f, 0.0f, 0.0f));
	m_Appearance.attachMaterial(&m_Material);
	m_Appearance.attachShader(&m_Shader);
	
	m_Mesh.setPrimitive(Mesh::TriangleList);
	m_Mesh.attachAppearance(&m_Appearance);	
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
SpringModel::~SpringModel() {
	m_Mesh.detachAppearance();
	m_Appearance.detachShader();
	m_Appearance.detachMaterial();
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
void SpringModel::setGeometry(unsigned int* geom,unsigned int size) {
	m_Indices.setSize(size);
	m_Indices.copy(geom, size);
	AutoPointer<IndexBuffer> ib = IndexBuffer::create(IndexBuffer::Index32, size, IndexBuffer::Static);
	m_Mesh.attachIndexBuffer(ib);
	ib->setBuffer(0, geom, size);
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
void SpringModel::onUpdate() {
	if (m_Springs == NULL)
		return;

	unsigned int count = m_Springs->getSize();
	VertexBuffer* vb = m_Mesh.getVertexBuffer();
	if ((vb == NULL) || (vb->getSize() != count)) {
		VertexFormat fmt;
		fmt.insert(Types::Vector3f, Attributes::Vertex);
		fmt.insert(Types::Vector3f, Attributes::Normal);
	
		vb = VertexBuffer::create(fmt, count, VertexBuffer::Stream);
		m_Mesh.attachVertexBuffer(vb);
		m_Vertices.setSize(count*2);
		vb->release();
	}
	
	Weight<float,3>* weight = &(*m_Springs)[0];
	Vector3<float>* vertex = &m_Vertices[0];
	Vector3<float>* vend = vertex + 2*count;
	
	// Init vertex list
	while (vertex < vend) {
		*vertex = weight->getPosition();
		vertex++;
		weight++;
		vertex->null();
		vertex++;
	}

	// Loop over all triangles
	unsigned int* index = &m_Indices[0];
	unsigned int* iend = index + m_Indices.getSize();
	while(index < iend) {
		unsigned int i0 = *index++;
		unsigned int i1 = *index++;
		unsigned int i2 = *index++;

		Vector3<float> normal = (m_Vertices[2*i1] - m_Vertices[2*i0]) % (m_Vertices[2*i2] - m_Vertices[2*i0]);
		m_Vertices[2*i0 + 1] += normal;
		m_Vertices[2*i1 + 1] += normal;
		m_Vertices[2*i2 + 1] += normal;
	}
	
	// Renormalize all normals and collect min/max info
	vertex = &m_Vertices[0];
	vend = vertex + 2*count;
	Vector3<float> vmin = m_Vertices[0];
	Vector3<float>vmax = m_Vertices[0];
	while (vertex < vend) {
		if (vertex->x <= vmin.x)
			vmin.x = vertex->x;
		else if (vertex->x >= vmax.x)
			vmax.x = vertex->x;
		if (vertex->y <= vmin.y)
			vmin.y = vertex->y;
		else if (vertex->y >= vmax.y)
			vmax.y = vertex->y;
		if (vertex->z <= vmin.z)
			vmin.z = vertex->z;
		else if (vertex->z >= vmax.z)
			vmax.z = vertex->z;
		vertex++;
		vertex->normalize();
		vertex++;
	}
	
	publishBounds(Bounds3<float>(AlignedBox3<float>(vmin,vmax)));
	
	// Copy vertices and normals into vertex buffer
	vb->setBuffer(0, &m_Vertices[0], count);
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
void SpringModel::onRender(const ParameterBuffer* pbuf) const {
	ASSERT(pbuf != NULL);
	MatrixStack* ms = getContext()->getMatrixStack();
	ms->selectMode(MatrixStack::ModelView);
	ms->set(*(const Matrix4<float>*)pbuf->getBuffer());
	m_Mesh.render();
}

}; };

