/* 
  Magnum 
  Copyright (C) 2002-2004 Kaya Memisoglu. All Rights Reserved.

  This software is provided 'as-is', without any express or implied
  warranty.  In no event will the authors be held liable for any damages
  arising from the use of this software.

  Permission is granted to anyone to use this software for any purpose,
  including commercial applications, and to alter it and redistribute it
  freely, under the terms in LICENSE.TXT.

  Kaya Memisoglu  (k.memisoglu@dimajix.de)
*/

#include "math.matrix.h"
#include "math.vector3.h"
#include "io.exception.h"
#include "io.inputstream.h"
#include "io.sysfile.h"
#include "graphics.translator.h"
#include "graphics3d.translator.3ds.h"
#include "graphics3d.indexmesh.h"
#include "graphics3d.colourmaterial.h"
#include "graphics3d.appearance.h"
#include "graphics3d.indexbuffer.h"
#include "graphics3d.vertexbuffer.h"
#include "graphics3d.staticmodel.h"
#include "graphics3d.texture2d.h"
#include "graphics3d.textureshader.h"
#include "graphics3d.plainshader.h"
#include "graphics3d.pixelprogram.h"
#include "graphics3d.vertexprogram.h"


#include "object.inl"
#include "object.pointer.inl"
#include "object.exception.inl"
#include "base.string.inl"
#include "math.matrix.inl"
#include "math.vector3.inl"
#include "io.inputstream.inl"
#include "system.thread.inl"
#include "container.map.inl"
#include "container.stack.inl"
#include "container.list.inl"
#include "container.allocator.inl"
#include "container.array.inl"
#include "graphics.colour.inl"
#include "graphics3d.mesh.inl"
#include "graphics3d.model.inl"
#include "graphics3d.vertexformat.inl"
#include "graphics3d.vertexmesh.inl"
#include "graphics3d.indexmesh.inl"
#include "graphics3d.material.inl"
#include "graphics3d.colourmaterial.inl"
#include "graphics3d.appearance.inl"
#include "graphics3d.indexbuffer.inl"
#include "graphics3d.vertexbuffer.inl"
#include "graphics3d.staticmodel.inl"
#include "graphics3d.context.inl"
#include "graphics3d.texture.inl"
#include "graphics3d.texture2d.inl"
#include "graphics3d.shader.inl"
#include "graphics3d.pixelprogram.inl"
#include "graphics3d.vertexprogram.inl"
#include "graphics3d.textureshader.inl"
#include "graphics3d.plainshader.inl"
#include "graphics3d.resource.inl"
#include "graphics3d.constant.inl"
#include "graphics3d.parameterformat.inl"



namespace magnum { namespace graphics3d {
using magnum::object::Exception;
using magnum::graphics::Image;
using magnum::math::Vector3;
using magnum::io::File;
using magnum::io::SystemFile;

// Each 3DS data-chunk starts with a 6 byte header.
// The first item in the header is a 2 byte (word) id-number.
// After that follows a dword wich gives the size of
// the data-chunk including the header. The size can be used
// as an relative offset to the next chunk.

// tab 4
enum {
	CHUNK_FLOAT_RGB    			= 0x0010,
	CHUNK_BYTE_RGB     			= 0x0011,
	CHUNK_INT_PERCENTAGE		= 0x0030, 
	CHUNK_FLOAT_PERCENTAGE		= 0x0031, 
//  CHUNK_RBGB2     			= 0x0012,	// ?? NOT HLS.
	CHUNK_MAIN      			= 0x4D4D,
     CHUNK_OBJMESH				= 0x3D3D,
      CHUNK_BKGCOLOR			= 0x1200,
	  CHUNK_AMBCOLOR  			= 0x2100,
	  CHUNK_OBJBLOCK  			= 0x4000,
	   CHUNK_TRIMESH   			= 0x4100,
		CHUNK_VERTLIST			= 0x4110,
		CHUNK_FACELIST			= 0x4120,
		CHUNK_FACEMAT			= 0x4130,
		CHUNK_MAPLIST			= 0x4140,
		CHUNK_SMOOLIST			= 0x4150,
		CHUNK_TRMATRIX			= 0x4160,
	   CHUNK_LIGHT     			= 0x4600,
		CHUNK_SPOTLIGHT			= 0x4610,
	   CHUNK_CAMERA    			= 0x4700,
	 CHUNK_MATERIAL  			= 0xAFFF,
	  CHUNK_MAT_NAME   			= 0xA000,
	  CHUNK_MAT_AMBIENT   		= 0xA010,
	  CHUNK_MAT_DIFFUSE   		= 0xA020,
	  CHUNK_MAT_SPECULAR  		= 0xA030,
	  CHUNK_MAT_SHININESS       = 0xA040,
	  CHUNK_MAT_SHIN2PCT        = 0xA041,
	  CHUNK_MAT_TRANSPARENCY    = 0xA050,
	  CHUNK_MAT_SELF_ILLUM		= 0xA080,
	  CHUNK_MAT_TWO_SIDE		= 0xA081,
	  CHUNK_MAT_DECAL			= 0xA082,
	  CHUNK_MAT_ADDITIVE		= 0xA083,
	  CHUNK_MAT_WIRE			= 0xA085,
	  CHUNK_MAT_TEXTURE   		= 0xA200,
	  CHUNK_MAT_BUMPMAP   		= 0xA230,
	  CHUNK_MAT_MAPFILE   		= 0xA300,
	  CHUNK_MAT_MAP_TILING      = 0xA351,
	  CHUNK_MAT_MAP_TEXBLUR     = 0xA353,
	  CHUNK_MAT_MAP_USCALE      = 0xA354,
	  CHUNK_MAT_MAP_VSCALE      = 0xA356,
	  CHUNK_MAT_MAP_UOFFSET     = 0xA358,
	  CHUNK_MAT_MAP_VOFFSET     = 0xA35A,
	  CHUNK_MAT_MAP_ANG         = 0xA35C,
	  CHUNK_MAT_MAP_COL1        = 0xA360,
	  CHUNK_MAT_MAP_COL2        = 0xA362,
	  CHUNK_MAT_MAP_RCOL        = 0xA364,
	  CHUNK_MAT_MAP_GCOL        = 0xA366,
	  CHUNK_MAT_MAP_BCOL        = 0xA368,			
	 CHUNK_KEYFRAMER 			= 0xB000,
	  CHUNK_FRAMES				= 0xB008
};


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
void Translator3DS::descendChunk() {
	Chunk c;
	c.m_Id = m_ChunkId;
	c.m_Len = m_ChunkLeft;
	m_ChunkStack.push(c);

	m_ChunkLeft = m_ChunkLen;
	m_ChunkLen = 0;
	m_ChunkId = 0;
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
bool Translator3DS::nextChunk() {
	Chunk c;
	long s;

	if (m_ChunkLeft == 0)
		return false;
	if (m_ChunkLeft < 6)
		throw magnum::io::TranslatorFormatException();

	if (m_ChunkLen > 0)
		m_In->skip(m_ChunkLen);

	s = m_In->read(&c, 6);
	m_ChunkId = c.m_Id;
	m_ChunkLen = c.m_Len-6;
	m_ChunkLeft -= c.m_Len;

	if ((s != 6) || (m_ChunkLen < 0) || (m_ChunkLeft < 0))
		throw magnum::io::TranslatorFormatException();

	return true;
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
void Translator3DS::ascendChunk() {
	m_In->skip(m_ChunkLeft + m_ChunkLen);

	Chunk c;
	c = m_ChunkStack.pop();

	m_ChunkId = c.m_Id;
	m_ChunkLen = 0;
	m_ChunkLeft = c.m_Len;
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
inline long Translator3DS::readChunk(void* dst, long size) {
	long ret;
	ret = m_In->read(dst, size);
	m_ChunkLen -= ret;
	if ((m_ChunkLen < 0) || (ret != size))
		throw magnum::io::TranslatorFormatException();
	return ret;
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
inline long Translator3DS::getChunkSize() {
	return m_ChunkLen;
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
inline word Translator3DS::getChunkId() {
	return m_ChunkId;
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
Colour<float> Translator3DS::readColour() {
	int size;

	descendChunk();
	while(nextChunk()) {
		switch(getChunkId()) {
			case CHUNK_BYTE_RGB:
				size = getChunkSize();
				byte bcol[3];
				if (size <= 3)
					readChunk(bcol, size);
				else
					readChunk(bcol, 3);
				ascendChunk();
				return Colour<float>((float)bcol[0]/255.0f, (float)bcol[1]/255.0f, (float)bcol[2]/255.0f);
			case CHUNK_FLOAT_RGB: 
				size = getChunkSize();
				float fcol[3];
				if (size <= 12)
					readChunk(fcol, size);
				else
					readChunk(fcol, 12);
				ascendChunk();
				return Colour<float>(fcol[0],fcol[1], fcol[2]);
			default: break;
		}
	}
	ascendChunk();

	return Colour<float>(0,0,0);
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
float Translator3DS::readScalar() {
	int size;

	descendChunk();
	while(nextChunk()) {
		switch(getChunkId()) {
			case CHUNK_INT_PERCENTAGE:
				size = getChunkSize();
				word w;
				readChunk(&w, sizeof(w));
				ascendChunk();
				return (float)w / 100.0f;
			case CHUNK_FLOAT_PERCENTAGE:
				size = getChunkSize();
				float f;
				readChunk(&f, sizeof(f));
				ascendChunk();
				return (float)f / 100.0f;
			default: break;
		}
	}
	ascendChunk();

	return 0.0f;
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
String Translator3DS::readString() {
	int size = getChunkSize();
	char* mapname = new char[size+1];
	readChunk(mapname, size);
	mapname[size] = 0;
	String s(mapname);
	delete[] mapname;
	return s;
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
void Translator3DS::readMaterialTexture() {
	String tname;

	descendChunk();
	while(nextChunk()) {
		switch(getChunkId()) {
			case CHUNK_MAT_MAPFILE: tname = readString(); break;
			case CHUNK_FLOAT_PERCENTAGE:
			case CHUNK_INT_PERCENTAGE:
			case CHUNK_MAT_MAP_TILING:   
			case CHUNK_MAT_MAP_TEXBLUR:  
			case CHUNK_MAT_MAP_USCALE:   
			case CHUNK_MAT_MAP_VSCALE:   
			case CHUNK_MAT_MAP_UOFFSET:  
			case CHUNK_MAT_MAP_VOFFSET:  
			case CHUNK_MAT_MAP_ANG:      
			case CHUNK_MAT_MAP_COL1:     
			case CHUNK_MAT_MAP_COL2:     
			case CHUNK_MAT_MAP_RCOL:     
			case CHUNK_MAT_MAP_GCOL:     
			case CHUNK_MAT_MAP_BCOL:     	
			default: break;
		}
	}
	ascendChunk();

	try {
		Texture* texture = createTexture(tname);
		if (texture == NULL)
			return;

		texture->setFilter(Texture::LinearMipmapLinear, Texture::Linear);
		m_Appearance->attachTexture(texture);
		texture->release();
	} catch (Exception ex) {

	}
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
void Translator3DS::readMaterial() {
	m_Appearance = new Appearance();
	m_Material = new ColourMaterial();
	m_Appearance->attachMaterial(m_Material);
	m_Material->release();

	descendChunk();
	while(nextChunk()) {
		switch(getChunkId()) {
			case CHUNK_MAT_NAME: m_Appearance->setName(readString()); break;
			case CHUNK_MAT_AMBIENT: m_Material->setAmbient(readColour()); break;  
	  		case CHUNK_MAT_DIFFUSE: m_Material->setDiffuse(readColour()); break;   
	  		case CHUNK_MAT_SPECULAR: m_Material->setSpecular(readColour()); break;   
			case CHUNK_MAT_SHININESS: m_Material->setShininess(readScalar()); break;
			case CHUNK_MAT_SELF_ILLUM: m_Material->setEmission(readColour()); break;
	  		case CHUNK_MAT_TEXTURE: readMaterialTexture(); break;
	  		case CHUNK_MAT_BUMPMAP:
			case CHUNK_MAT_TRANSPARENCY:
			case CHUNK_MAT_TWO_SIDE:
  			case CHUNK_MAT_DECAL:
			case CHUNK_MAT_ADDITIVE:
			case CHUNK_MAT_WIRE:
			default: break;
		}
	}
	ascendChunk();

	if (m_Appearance->getTexture() != NULL)
		m_Appearance->attachShader(m_TextureShader);
	else
		m_Appearance->attachShader(m_PlainShader);

	m_Materials.insert(m_Appearance->getName(), m_Appearance);
	m_Appearance->release();
	m_Appearance = NULL;
	m_Material = NULL;
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
void Translator3DS::readVertexList() {
	word nv;
	readChunk(&nv, sizeof(nv));

	if (m_VertexCount == 0)
		m_VertexCount = nv;
	else
		m_VertexCount = min(m_VertexCount, (unsigned int)nv);

	m_VertexList = new float32[nv*3];
	readChunk(m_VertexList, sizeof(float32)*3*nv);
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
void Translator3DS::readFaceList() {
	word nv;
	readChunk(&nv, sizeof(nv));

	m_FaceCount = nv;
	m_FaceList = new word[nv*4];
	readChunk(m_FaceList, sizeof(word)*4*nv);

	descendChunk();
	while(nextChunk()) {
		switch(getChunkId()) {
			case CHUNK_FACEMAT:	 readFaceMaterial(); break;
			case CHUNK_SMOOLIST:
			default: break;
		}
	}
	ascendChunk();
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
void Translator3DS::readMapList() {
	word nv;
	readChunk(&nv, sizeof(nv));

	if (m_VertexCount == 0)
		m_VertexCount = nv;
	else
		m_VertexCount = min(m_VertexCount, (unsigned int)nv);

	m_MapList = new float32[2*nv];
	readChunk(m_MapList, sizeof(float32)*2*nv);
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
void Translator3DS::readFaceMaterial() {
	char name[64];
	long size = min(getChunkSize(), (long)63);

	// The material name is the first item
	int n=0;
	do {
		readChunk(&name[n++], 1);
		size--;
	} while((name[n-1] != '\0') && (size > 0));
	name[n-1] = '\0';

	word f;
	readChunk(&f, sizeof(f));
	if (f == 0)
		return;

	Appearance* app;
	if (m_Materials.contains(String(name))) {
		app = m_Materials[String(name)]; 
	} else {
		app = NULL;
	}

	FaceMaterial fapp;
	fapp.m_Appearance = app;
	fapp.m_FaceCount = f;
	fapp.m_FaceList = new word[f];
	readChunk(fapp.m_FaceList, f*sizeof(word));

	m_FaceMaterials.insert(fapp);
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
void Translator3DS::readTraMatrix() {
    float mtx[12];
	readChunk(&mtx, sizeof(float32)*12);
	m_Matrix[0][0] = mtx[0];
	m_Matrix[0][1] = mtx[1];
	m_Matrix[0][2] = mtx[2];
	m_Matrix[0][3] = mtx[9];
	m_Matrix[1][0] = mtx[3];
	m_Matrix[1][1] = mtx[4];
	m_Matrix[1][2] = mtx[5];
	m_Matrix[1][3] = mtx[10];
	m_Matrix[2][0] = mtx[6];
	m_Matrix[2][1] = mtx[7];
	m_Matrix[2][2] = mtx[8];
	m_Matrix[2][3] = mtx[11];
	m_Matrix[3][0] = 0;
	m_Matrix[3][1] = 0;
	m_Matrix[3][2] = 0;
	m_Matrix[3][3] = 1;
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
void Translator3DS::readTriMesh() {

	m_VertexBuffer = NULL;
	m_IndexBuffer = NULL;
	m_VertexList = NULL;
	m_FaceList = NULL;
	m_MapList = NULL;
	m_VertexCount = 0;
	m_FaceCount = 0;
	m_FaceMaterials.clear();

	descendChunk();
	while(nextChunk()) {
		switch(getChunkId()) {
			case CHUNK_VERTLIST: readVertexList(); break;
			case CHUNK_FACELIST: readFaceList(); break;
			case CHUNK_MAPLIST:  readMapList(); break;
			case CHUNK_TRMATRIX: readTraMatrix(); break;
			case CHUNK_FACEMAT:	 readFaceMaterial(); break;
			case CHUNK_SMOOLIST:
			default: break;
		}
	}
	ascendChunk();

	if ((m_VertexList == NULL) || (m_FaceList == NULL)) {
		return;
	}

	createVertexBuffer();
	createIndexBuffer();
	createMeshes();

	m_VertexBuffer->release();
	m_IndexBuffer->release();
}



/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
void Translator3DS::readObjectBlock() {
	char name[64];
	long size = min(getChunkSize(), (long)63);

	// The object name is the first item
	int n=0;
	do {
		readChunk(&name[n++], 1);
		size--;
	} while((name[n-1] != '\0') && (size > 0));
	name[n-1] = '\0';

	descendChunk();
	while(nextChunk()) {
		switch(getChunkId()) {
			case CHUNK_TRIMESH:	readTriMesh(); break;
			//case CHUNK_LIGHT:
			//case CHUNK_CAMERA:
			default: break;
		}
	}
	ascendChunk();
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
void Translator3DS::readMeshWorld() {
	// Allocate a new Model
	if (m_Model == NULL)
		m_Model = new StaticModel();

	descendChunk();
	while(nextChunk()) {
		switch(getChunkId()) {
			case CHUNK_OBJBLOCK: readObjectBlock(); break;
			case CHUNK_MATERIAL: readMaterial(); break;
			//case CHUNK_AMBCOLOR:
			//case CHUNK_BKGCOLOR:
			default: break;
		}
	}
	ascendChunk();

	// finish Model
	m_Model->finish();
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
void Translator3DS::createIndexBuffer() {
	unsigned int fc = 0;

	for (unsigned int i = 0; i < m_FaceMaterials.getSize(); i++)
		fc += m_FaceMaterials[i].m_FaceCount;

	m_IndexBuffer = IndexBuffer::create(IndexBuffer::Index16, fc*3, IndexBuffer::Static);
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
void Translator3DS::createMeshes() {
	// Create meshes (for each material one mesh)
	unsigned int range = 0;

	for (unsigned int i = 0; i < m_FaceMaterials.getSize(); i++) {
		Appearance* app = m_FaceMaterials[i].m_Appearance;
		unsigned int fc = m_FaceMaterials[i].m_FaceCount;
		word* fl = m_FaceMaterials[i].m_FaceList;
		if (fc == 0)
			continue;

		IndexMesh* mesh = new IndexMesh();
		mesh->setPrimitive(Mesh::TriangleList);
		mesh->attachVertexBuffer(m_VertexBuffer);
		mesh->attachIndexBuffer(m_IndexBuffer);
		mesh->setRange(range, 3*fc);

		if (app != NULL)
			mesh->attachAppearance(app);

		// Create Index Buffer
		word* wptr = (word *) m_IndexBuffer->lock(IndexBuffer::WriteOnly);
		for (unsigned int j = 0; j < fc; j++) {
			wptr[3*j + range] = m_FaceList[4*fl[j]];
			wptr[3*j + 1 + range] = m_FaceList[4*fl[j] + 1];
			wptr[3*j + 2 + range] = m_FaceList[4*fl[j] + 2];
		}
		m_IndexBuffer->unlock();

		delete[] fl;
		m_Model->insertMesh(mesh);
		mesh->release();

		range += 3*fc;
	}
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
void Translator3DS::createVertexBuffer() {
	// Create Vertex Buffer
	unsigned int i;
	VertexFormat fmt;
	VertexBuffer* vb;
	
	fmt.set(0, Types::Float3, Attributes::Vertex);
	fmt.set(1, Types::Float3, Attributes::Normal);
	if (m_MapList != NULL)
		fmt.set(2, Types::Float2, Attributes::TextureAll);
	vb = VertexBuffer::create(fmt, m_VertexCount, VertexBuffer::Static);

	// Copy Vertex data;
	void* vbptr;
	int vbstride = vb->getStride();
	float *fptr;
	vbptr = vb->lock(VertexBuffer::Discard);

	// Copy Vertex Vectors
	fptr = (float *)((byte *)vbptr + vb->getOffset(0));
	for (i = 0; i < m_VertexCount; i++) {
		fptr[0] = m_VertexList[i*3];
		fptr[1] = m_VertexList[i*3+1];
		fptr[2] = m_VertexList[i*3+2];
		fptr = (float *)(((byte *)(void *)fptr) + vbstride);
	}

	// Copy Texture Mappings
	if (m_MapList != NULL) {
		fptr = (float *)((byte *)vbptr + vb->getOffset(2));
		for (i = 0; i < m_VertexCount; i++) {
			fptr[0] = m_MapList[i*2];
			fptr[1] = 1.0f - m_MapList[i*2+1];
			fptr = (float *)(((byte *)(void *)fptr) + vbstride);
		}
	}

	// Create Normals
	fptr = (float *)((byte *)vbptr + vb->getOffset(1));
	for (i = 0; i < m_VertexCount; i++) {
		fptr[0] = 0.0f;
		fptr[1] = 0.0f;
		fptr[2] = 0.0f;
		fptr = (float *)(((byte *)(void *)fptr) + vbstride);
	}
	fptr = (float *)((byte *)vbptr + vb->getOffset(1));
	for (i = 0; i < m_FaceCount; i++) {
		int vi1,vi2,vi3;
		vi1 = m_FaceList[4*i];
		vi2 = m_FaceList[4*i+1];
		vi3 = m_FaceList[4*i+2];

		Vector3<float> v1 = Vector3<float>(&m_VertexList[vi1*3]);
		Vector3<float> v2 = Vector3<float>(&m_VertexList[vi2*3]);
		Vector3<float> v3 = Vector3<float>(&m_VertexList[vi3*3]);
		Vector3<float> normal = (v2-v1) % (v3-v1);

		float *vp;
		vp = (float *)(((byte *)(void *)fptr) + vi1*vbstride);
		vp[0] += normal.x;
		vp[1] += normal.y;
		vp[2] += normal.z;
		vp = (float *)(((byte *)(void *)fptr) + vi2*vbstride);
		vp[0] += normal.x;
		vp[1] += normal.y;
		vp[2] += normal.z;
		vp = (float *)(((byte *)(void *)fptr) + vi3*vbstride);
		vp[0] += normal.x;
		vp[1] += normal.y;
		vp[2] += normal.z;
	}
	fptr = (float *)((byte *)vbptr + vb->getOffset(1));
	for (i = 0; i < m_VertexCount; i++) {
		Vector3<float> v = Vector3<float>(fptr);
		v = v / v.length();
		fptr[0] = v.x;
		fptr[1] = v.y;
		fptr[2] = v.z;
		fptr = (float *)(((byte *)(void *)fptr) + vbstride);
	}

	vb->unlock();
	m_VertexBuffer = vb;
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
Model* Translator3DS::onDecode(InputStream* in,long size) {
	if (size == Autosize)
		size = in->available();

	// Init some vars...
	m_In = in;
	m_Model = NULL;
	m_ChunkLen = 0;
	m_ChunkLeft = size;
	m_ChunkStack.clear();
	m_PlainShader = new PlainShader();
	m_TextureShader = new TextureShader();

	// Read main chunk
	if (nextChunk() == 0)
		throw magnum::io::TranslatorFormatException();
	if (getChunkId() != CHUNK_MAIN)
		throw magnum::io::TranslatorFormatException();

	descendChunk();
	while(nextChunk()) {
		switch(getChunkId()) {
			case CHUNK_OBJMESH: readMeshWorld(); break;
			case CHUNK_MATERIAL: readMaterial(); break;
			//case CHUNK_KEYFRAMER:
			default: break;
		}
	}
	ascendChunk();

	m_TextureShader->release();
	m_PlainShader->release();

	m_FaceMaterials.clear();
	m_Materials.clear();

	return m_Model;
}



/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
long Translator3DS::onEncode(OutputStream* ,const Model* ) {
	return 0;
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
bool Translator3DS::onIdentify(File* in) {
	word id;

	try {
		unsigned long oldpos;
		oldpos = in->getPosition();	
		in->read(&id, sizeof(id));
		in->setPosition(oldpos, File::Absolute);
	} catch (...) {
		return false;
	}

	if (id != CHUNK_MAIN)
		return false;

	return true;
}

}; };
