/* 
  Magnum 
  Copyright (C) 2002-2004 Kaya Memisoglu. All Rights Reserved.

  This software is provided 'as-is', without any express or implied
  warranty.  In no event will the authors be held liable for any damages
  arising from the use of this software.

  Permission is granted to anyone to use this software for any purpose,
  including commercial applications, and to alter it and redistribute it
  freely, under the terms in LICENSE.TXT.

  Kaya Memisoglu  (k.memisoglu@dimajix.de)
*/

#include "graphics3d.context.h"
#include "graphics3d.vertexbuffer.h"

#include "object.inl"
#include "base.string.inl"
#include "system.thread.inl"
#include "system.assembler.inl"
#include "graphics3d.context.inl"
#include "graphics3d.vertexbuffer.inl"
#include "graphics3d.vertexformat.inl"
#include "graphics3d.resource.inl"


namespace magnum { namespace graphics3d {

const int VertexBuffer::g_OglAcces[] = {
	0,
	GL_READ_ONLY,	// ReadOnly = 1,
	GL_WRITE_ONLY,	// WriteOnly = 2,
	GL_READ_WRITE,	// ReadWrite = 3,
	GL_WRITE_ONLY	// Discard = 4
};

const int VertexBuffer::g_OglUsage[] = {
	0,
	GL_STATIC_DRAW,  // Static = 1,
	GL_DYNAMIC_DRAW, // Dynamic = 2,
	GL_STREAM_DRAW,	 // Stream = 3
	GL_DYNAMIC_DRAW, // System = 4,
};
	

/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
VertexBuffer* VertexBuffer::create(const VertexFormat& fmt, unsigned int count, CreateFlag cf) {
	VertexBuffer* vb = new VertexBuffer(fmt);
	if (vb == NULL)
		return NULL;

	vb->m_Size = count;
	vb->m_OglUsage = g_OglUsage[cf];

	// Try to create Buffer Object
	if ((glBindBufferARB != NULL) && (cf != System)) {
		glGetError();
		glGenBuffersARB(1, &vb->m_OglBuffer);
		if (glGetError() != GL_NO_ERROR) {
			vb->m_OglBuffer = NULL;
			delete vb;
			return NULL;
		}
			
		glBindBufferARB(GL_ARRAY_BUFFER, vb->m_OglBuffer);
		glBufferDataARB(GL_ARRAY_BUFFER, count*fmt.getStride(), NULL, vb->m_OglUsage);
		if (glGetError() != GL_NO_ERROR) {
			delete vb;
			return NULL;
		}
	} else {
		// Fall back to simple memory buffer
		vb->m_Buffer = (void *)new byte[count * fmt.getStride()];
		if (vb->m_Buffer == NULL) {
			delete vb;
			return NULL;
		}
	}

	return vb;
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
VertexBuffer::VertexBuffer(const VertexFormat& fmt) : m_Format(fmt) {
	unsigned int oglSize, oglType, attribs, offset;
	m_Locked = false;
	m_Buffer = NULL;
	m_Size = 0;
	m_Attributes = 0;
	m_OglBuffer = 0;
	m_OglUsage = 0;

	memset(m_Mapping, 0, sizeof(m_Mapping));

	for (unsigned int i = 0; i < m_Format.getSize(); i++) {
		if ((m_Format[i].m_Attribute == 0) || (m_Format[i].m_Type == 0))
			continue;

		offset = m_Format[i].m_Offset;
		attribs = m_Format[i].m_Attribute;
		oglSize = (m_Format[i].m_Type >> Types::CountShift) & Types::CountMask;
		switch((m_Format[i].m_Type >> Types::TypeShift ) & Types::TypeMask) {
			case (Types::Byte >> Types::TypeShift) & Types::TypeMask:
				oglType = GL_BYTE;
				break;
			case (Types::UByte >> Types::TypeShift) & Types::TypeMask:
				oglType = GL_UNSIGNED_BYTE;
				break;
			case (Types::Short >> Types::TypeShift) & Types::TypeMask:
				oglType = GL_SHORT;
				break;
			case (Types::UShort >> Types::TypeShift) & Types::TypeMask:
				oglType = GL_UNSIGNED_SHORT;
				break;
			case (Types::Int >> Types::TypeShift) & Types::TypeMask:
				oglType = GL_INT;
				break;
			case (Types::UInt >> Types::TypeShift) & Types::TypeMask:
				oglType = GL_UNSIGNED_INT;
				break;
			case (Types::Float >> Types::TypeShift) & Types::TypeMask:
				oglType = GL_FLOAT;
				break;
			case (Types::Double >> Types::TypeShift) & Types::TypeMask:
				oglType = GL_DOUBLE;
				break;
		}

		((VertexFormat::Component*)&m_Format[i])->m_OglType = oglType;
		((VertexFormat::Component*)&m_Format[i])->m_OglSize = oglSize;

		m_Attributes |= attribs;
		for (unsigned int j = 0; j < Attributes::MaxAttributes; j++) {
			int idx = attribs & (1 << j);
			if (idx != 0) {
				m_Mapping[j].m_OglType = oglType;
				m_Mapping[j].m_OglSize = oglSize;
				m_Mapping[j].m_Offset = offset;
			}
		}
	}
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
VertexBuffer::~VertexBuffer() {
	if (m_OglBuffer != 0) {
		glDeleteBuffersARB(1, &m_OglBuffer);
	} else {
		if (m_Buffer != NULL)
			delete[] (byte *)m_Buffer;
	}
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
void VertexBuffer::setBuffer(unsigned int offset, const void* data, unsigned int size) {
	ASSERT(offset < m_Size);
	ASSERT(offset + size <= m_Size);
	ASSERT(!m_Locked);
	offset *= m_Format.getStride();
	size *= m_Format.getStride();
	if (m_OglBuffer) {
		glBindBufferARB(GL_ARRAY_BUFFER, m_OglBuffer);
		glBufferSubDataARB(GL_ARRAY_BUFFER, offset, size, data);
	} else {
		memcpy(((byte *)m_Buffer) + offset, data, size);
	}
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
void VertexBuffer::getBuffer(unsigned int offset, void* data, unsigned int size) const {
	ASSERT(offset < m_Size);
	ASSERT(offset + size <= m_Size);
	ASSERT(!m_Locked);
	offset *= m_Format.getStride();
	size *= m_Format.getStride();
	if (m_OglBuffer) {
		glBindBufferARB(GL_ARRAY_BUFFER, m_OglBuffer);
		glGetBufferSubDataARB(GL_ARRAY_BUFFER, offset, size, data);
	} else {
		memcpy(data, ((byte *)m_Buffer) + offset, size);
	}
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
void VertexBuffer::select(unsigned int flags) const {
	ASSERT((m_Buffer != NULL) || (m_OglBuffer != NULL));
	flags &= m_Attributes;

	// First set vertex buffer object - if needed
	byte* base = (byte *)bindBuffer();

	// Set Vertex Stuff
	if (flags & Attributes::Vertex) {
		ASSERT((m_Mapping[Attributes::VertexBit].m_OglSize >= 3) 
			&& (m_Mapping[Attributes::VertexBit].m_OglSize <= 4)); 
		glVertexPointer(m_Mapping[Attributes::VertexBit].m_OglSize, 
						m_Mapping[Attributes::VertexBit].m_OglType, 
						getStride(), 
						base + m_Mapping[Attributes::VertexBit].m_Offset);
	}

	// Set Normal Stuff
	if (flags & Attributes::Normal) {
		ASSERT(m_Mapping[Attributes::NormalBit].m_OglSize == 3);
		glNormalPointer(m_Mapping[Attributes::NormalBit].m_OglType, 
						getStride(), 
						base + m_Mapping[Attributes::NormalBit].m_Offset);
	}

	// Set Vertex Colour Stuff
	if (flags & Attributes::Colour) {
		ASSERT((m_Mapping[Attributes::ColourBit].m_OglSize == 3) 
			|| (m_Mapping[Attributes::ColourBit].m_OglSize == 4)); 
		glColorPointer(m_Mapping[Attributes::ColourBit].m_OglSize, 
						m_Mapping[Attributes::ColourBit].m_OglType, 
						getStride(), 
						base + m_Mapping[Attributes::ColourBit].m_Offset);
	}

	// Set secondary colour

	// Set weights
	if (flags & Attributes::Weights) {
		glWeightPointerARB(m_Mapping[Attributes::WeightsBit].m_OglSize, 
						m_Mapping[Attributes::WeightsBit].m_OglType, 
						getStride(), 
						base + m_Mapping[Attributes::WeightsBit].m_Offset);
	}

	// Set fog coords
	if (flags & Attributes::Fog) {
		ASSERT(m_Mapping[Attributes::FogBit].m_OglSize == 1);
		glFogCoordPointerEXT(m_Mapping[Attributes::FogBit].m_OglType, 
					getStride(), 
					base + m_Mapping[Attributes::FogBit].m_Offset);
	}

	// Set edge flags
	if (flags & Attributes::Edge) {
		ASSERT(m_Mapping[Attributes::EdgeBit].m_OglSize == 1);
		glEdgeFlagPointer(getStride(), base + m_Mapping[Attributes::ColourBit].m_Offset);
	}

	// Set Texture Coordinate Stuff
	for (int i = 0; i < Texture::MaxTextures; i++) {
		if (flags & (Attributes::Texture << i)) {
			ASSERT((m_Mapping[Attributes::TextureBit + i].m_OglSize >= 1) 
				|| (m_Mapping[Attributes::TextureBit + i].m_OglSize <= 4));
			glClientActiveTextureARB(GL_TEXTURE0_ARB + i);
			glTexCoordPointer(m_Mapping[Attributes::TextureBit + i].m_OglSize, 
							m_Mapping[Attributes::TextureBit + i].m_OglType, 
							getStride(), 
							base + m_Mapping[Attributes::TextureBit + i].m_Offset);
		}
	}
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
void VertexBuffer::compileSelect(Assembler& as, unsigned int flags) const {
	ASSERT((m_Buffer != NULL) || (m_OglBuffer != NULL));
	flags &= m_Attributes;

	// First set vertex buffer object - if needed
	byte* base;
	if (m_OglBuffer) {
		as.asmCall(glBindBufferARB, (GLenum)GL_ARRAY_BUFFER, (GLuint)m_OglBuffer);
		base = NULL;
	} else {
		if (glBindBufferARB)
			as.asmCall(glBindBufferARB, (GLenum)GL_ARRAY_BUFFER, (GLuint)NULL);
		base = (byte*)m_Buffer;
	}

	// Set Vertex Stuff
	if (flags & Attributes::Vertex) {
		ASSERT((m_Mapping[Attributes::VertexBit].m_OglSize >= 3) 
			&& (m_Mapping[Attributes::VertexBit].m_OglSize <= 4)); 
		as.asmCall(glVertexPointer, 
			(GLint)m_Mapping[Attributes::VertexBit].m_OglSize, 
			(GLenum)m_Mapping[Attributes::VertexBit].m_OglType, 
			(GLsizei)getStride(), 
			(const GLvoid*)(base + m_Mapping[Attributes::VertexBit].m_Offset));
	}

	// Set Normal Stuff
	if (flags & Attributes::Normal) {
		ASSERT(m_Mapping[Attributes::NormalBit].m_OglSize == 3);
		as.asmCall(glNormalPointer, 
			(GLenum)m_Mapping[Attributes::NormalBit].m_OglType, 
			(GLsizei)getStride(), 
			(const GLvoid*)(base + m_Mapping[Attributes::NormalBit].m_Offset));
	}

	// Set Vertex Colour Stuff
	if (flags & Attributes::Colour) {
		ASSERT((m_Mapping[Attributes::ColourBit].m_OglSize == 3) 
			|| (m_Mapping[Attributes::ColourBit].m_OglSize == 4)); 
		as.asmCall(glColorPointer, 
			(GLint)m_Mapping[Attributes::ColourBit].m_OglSize, 
			(GLenum)m_Mapping[Attributes::ColourBit].m_OglType, 
			(GLsizei)getStride(), 
			(const GLvoid*)(base + m_Mapping[Attributes::ColourBit].m_Offset));
	}

	// Set secondary colour

	// Set weights
	if (flags & Attributes::Weights) {
		as.asmCall(glWeightPointerARB, 
			(GLint)m_Mapping[Attributes::WeightsBit].m_OglSize, 
			(GLenum)m_Mapping[Attributes::WeightsBit].m_OglType, 
			(GLsizei)getStride(), 
			(const GLvoid*)(base + m_Mapping[Attributes::WeightsBit].m_Offset));
	}

	// Set fog coords
	if (flags & Attributes::Fog) {
		ASSERT(m_Mapping[Attributes::FogBit].m_OglSize == 1);
		as.asmCall(glFogCoordPointerEXT, 
			(GLenum)m_Mapping[Attributes::FogBit].m_OglType, 
			(GLsizei)getStride(), 
			(const GLvoid*)(base + m_Mapping[Attributes::FogBit].m_Offset));
	}

	// Set edge flags
	if (flags & Attributes::Edge) {
		ASSERT(m_Mapping[Attributes::ColourBit].m_OglSize == 1);
		as.asmCall(glEdgeFlagPointer, 
			(GLsizei)getStride(), 
			(const GLvoid*)(base + m_Mapping[Attributes::ColourBit].m_Offset));
	}

	// Set Texture Coordinate Stuff
	for (int i = 0; i < Texture::MaxTextures; i++) {
		if (flags & (Attributes::Texture << i)) {
			ASSERT((m_Mapping[Attributes::TextureBit + i].m_OglSize >= 1) 
				|| (m_Mapping[Attributes::TextureBit + i].m_OglSize <= 4));
			as.asmCall(glClientActiveTextureARB, 
				(GLenum)(GL_TEXTURE0_ARB + i));
			as.asmCall(glTexCoordPointer, 
				(GLint)m_Mapping[Attributes::TextureBit + i].m_OglSize, 
				(GLenum)m_Mapping[Attributes::TextureBit + i].m_OglType, 
				(GLsizei)getStride(), 
				(const GLvoid*)(base + m_Mapping[Attributes::TextureBit + i].m_Offset));
		}
	}
}


}; };
