/* 
  Magnum 
  Copyright (C) 2002-2004 Kaya Memisoglu. All Rights Reserved.

  This software is provided 'as-is', without any express or implied
  warranty.  In no event will the authors be held liable for any damages
  arising from the use of this software.

  Permission is granted to anyone to use this software for any purpose,
  including commercial applications, and to alter it and redistribute it
  freely, under the terms in LICENSE.TXT.

  Kaya Memisoglu  (k.memisoglu@dimajix.de)
*/

#include "graphics3d.context.h"
#include "graphics3d.vertexformat.h"
#include "graphics3d.vertexconverter.h"

#include "graphics3d.context.inl"
#include "graphics3d.vertexconverter.inl"
#include "graphics3d.vertexformat.inl"
#include "graphics3d.vertexbuffer.inl"


namespace magnum { namespace graphics3d {

/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
VertexConverter::VertexConverter() {
	memset(m_InputAttribute, 0, sizeof(m_InputAttribute));
	memset(m_OutputAttribute, 0, sizeof(m_OutputAttribute));
	memset(m_Input, 0, sizeof(m_Input));
	memset(m_Output, 0, sizeof(m_Output));
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
VertexConverter::~VertexConverter() {
	detachInput();
	detachOutput();
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
bool VertexConverter::attachInput(const VertexBuffer* buf) {
	ASSERT(buf != NULL);
	for (unsigned int i = 0; i < VertexBuffer::MaxBuffers; i++) {
		if (m_Input[i] == NULL) {
			m_Input[i] = (VertexBuffer*)buf;
			m_Input[i]->refer();
			addAttributes(buf, i, m_InputAttribute);
			return true;
		}
	}

	return false;
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
bool VertexConverter::detachInput(const VertexBuffer* buf) {
	ASSERT(buf != NULL);
	for (unsigned int i = 0; i < VertexBuffer::MaxBuffers; i++) {
		if (m_Input[i] == buf) {
			m_Input[i]->release();
			m_Input[i] = NULL;
			removeAttributes(buf, m_InputAttribute);
			return true;
		}
	}

	return false;
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
void VertexConverter::detachInput() {
	for (unsigned int i = 0; i < VertexBuffer::MaxBuffers; i++) {
		if (m_Input[i] != NULL) {
			m_Input[i]->release();
		}
	}
	memset(m_InputAttribute, 0, sizeof(m_InputAttribute));
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
bool VertexConverter::attachOutput(VertexBuffer* buf) {
	ASSERT(buf != NULL);
	for (unsigned int i = 0; i < VertexBuffer::MaxBuffers; i++) {
		if (m_Output[i] == NULL) {
			m_Output[i] = buf;
			buf->refer();
			addAttributes(buf, i, m_OutputAttribute);
			return true;
		}
	}

	return false;
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
bool VertexConverter::detachOutput(VertexBuffer* buf) {
	ASSERT(buf != NULL);
	for (unsigned int i = 0; i < VertexBuffer::MaxBuffers; i++) {
		if (m_Output[i] == buf) {
			m_Output[i] = NULL;
			buf->release();
			removeAttributes(buf, m_OutputAttribute);
			return true;
		}
	}

	return false;
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
void VertexConverter::detachOutput() {
	for (unsigned int i = 0; i < VertexBuffer::MaxBuffers; i++) {
		if (m_Output[i] != NULL) {
			m_Output[i]->release();
		}
	}
	memset(m_OutputAttribute, 0, sizeof(m_OutputAttribute));
}
 

/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
bool VertexConverter::convert() {
	unsigned int size = 0;
	unsigned int i,j,k;

	// Check buffer sizes
	for (i = 0; i < VertexBuffer::MaxBuffers; i++) {
		if (m_Input[i] != NULL) {
			if (size == 0) {
				size = m_Input[i]->getSize();
			}
			ASSERT(size == m_Input[i]->getSize());
		}
		if (m_Output[i] != NULL) {
			if (size == 0) {
				size = m_Output[i]->getSize();
			}
			ASSERT(size == m_Output[i]->getSize());
		}
	}

	// Lock all buffers
	unsigned int istride[VertexBuffer::MaxBuffers];
	unsigned int ostride[VertexBuffer::MaxBuffers];
	byte* ibuf[VertexBuffer::MaxBuffers];
	byte* obuf[VertexBuffer::MaxBuffers];
	for (i = 0; i < VertexBuffer::MaxBuffers; i++) {
		if (m_Input[i] != NULL) {
			istride[i] = m_Input[i]->getStride();
			ibuf[i] = (byte*)m_Input[i]->lock(VertexBuffer::ReadOnly);
		} else {
			istride[i] = 0;
			ibuf[i] = 0;
		}

		if (m_Output[i] != NULL) {
			ostride[i] = m_Output[i]->getStride();
			obuf[i] = (byte*)m_Output[i]->lock(VertexBuffer::WriteOnly);
		} else {
			ostride[i] = 0;
			obuf[i] = 0;
		}
	}

	// Start conversion
	double data[8];

	float* fptr;
	double* dptr;
	int8* bptr;
	int16* sptr;
	int32* iptr;

	unsigned int type;
	unsigned int count;
	unsigned int offset;
	for (i = 0; i < size; i++) {
		// Loop over all Attributes
		for (j = 0; j < Attributes::MaxAttributes; j++) {
			if (m_InputAttribute[j].m_Buffer == NULL)
				continue;
			if (m_OutputAttribute[j].m_Buffer == NULL)
				continue;

			type = m_InputAttribute[j].m_TypeBase;
			count = m_InputAttribute[j].m_TypeCount;
			offset = m_InputAttribute[j].m_Offset;
			bptr = (int8*)ibuf[m_InputAttribute[j].m_BufferIdx] + offset;

			for (k = 0; k < 8; k++)
				data[k] = 0;
			
			// Read value
			switch(type) {
				case ((Types::Byte >> Types::TypeShift) & Types::TypeMask):
					for (k = 0; k < count; k++) {
						data[k] = (double)(int8)bptr[k];
					}
					break;
				case ((Types::UByte >> Types::TypeShift) & Types::TypeMask):
					for (k = 0; k < count; k++) {
						data[k] = (double)(uint8)bptr[k];
					}
					break;
				case ((Types::Short >> Types::TypeShift) & Types::TypeMask):
					sptr = (int16*)bptr;
					for (k = 0; k < count; k++) {
						data[k] = (double)(int16)sptr[k];
					}
					break;
				case ((Types::UShort >> Types::TypeShift) & Types::TypeMask):
					sptr = (int16*)bptr;
					for (k = 0; k < count; k++) {
						data[k] = (double)(uint16)sptr[k];
					}
					break;
				case ((Types::Int >> Types::TypeShift) & Types::TypeMask):
					iptr = (int32*)bptr;
					for (k = 0; k < count; k++) {
						data[k] = (double)(int32)iptr[k];
					}
					break;
				case ((Types::UInt >> Types::TypeShift) & Types::TypeMask):
					iptr = (int32*)bptr;
					for (k = 0; k < count; k++) {
						data[k] = (double)(uint32)iptr[k];
					}
					break;
				case ((Types::Float >> Types::TypeShift) & Types::TypeMask):
					fptr = (float*)bptr;
					for (k = 0; k < count; k++) {
						data[k] = (double)(float)fptr[k];
					}
					break;
				case ((Types::Double >> Types::TypeShift) & Types::TypeMask):
					dptr = (double*)bptr;
					for (k = 0; k < count; k++) {
						data[k] = (double)dptr[k];
					}
					break;
			}


			type = m_OutputAttribute[j].m_TypeBase;
			count = m_OutputAttribute[j].m_TypeCount;
			offset = m_OutputAttribute[j].m_Offset;
			bptr = (int8*)obuf[m_OutputAttribute[j].m_BufferIdx] + offset;

			// Write value
			switch(type) {
				case ((Types::Byte >> Types::TypeShift) & Types::TypeMask):
					for (k = 0; k < count; k++) {
						bptr[k] = (int8)data[k];
					}
					break;
				case ((Types::UByte >> Types::TypeShift) & Types::TypeMask):
					for (k = 0; k < count; k++) {
						bptr[k] = (int8)(uint8)data[k];
					}
					break;
				case ((Types::Short >> Types::TypeShift) & Types::TypeMask):
					sptr = (int16*)bptr;
					for (k = 0; k < count; k++) {
						sptr[k] = (int16)data[k];
					}
					break;
				case ((Types::UShort >> Types::TypeShift) & Types::TypeMask):
					sptr = (int16*)bptr;
					for (k = 0; k < count; k++) {
						sptr[k] = (int16)(uint16)data[k];
					}
					break;
				case ((Types::Int >> Types::TypeShift) & Types::TypeMask):
					iptr = (int32*)bptr;
					for (k = 0; k < count; k++) {
						iptr[k] = (int32)data[k];
					}
					break;
				case ((Types::UInt >> Types::TypeShift) & Types::TypeMask):
					iptr = (int32*)bptr;
					for (k = 0; k < count; k++) {
						iptr[k] = (int32)(uint32)data[k];
					}
					break;
				case ((Types::Float >> Types::TypeShift) & Types::TypeMask):
					fptr = (float*)bptr;
					for (k = 0; k < count; k++) {
						fptr[k] = (float)data[k];
					}
					break;
				case ((Types::Double >> Types::TypeShift) & Types::TypeMask):
					dptr = (double*)bptr;
					for (k = 0; k < count; k++) {
						dptr[k] = (double)data[k];
					}
					break;
			}
		}

		// Advance one step in all vertex buffers
		for (j = 0; j < VertexBuffer::MaxBuffers; j++) {
			ibuf[j] = ibuf[j] + istride[j];
			obuf[j] = obuf[j] + ostride[j];
		}
	}

	// Unlock all buffers
	for (i = 0; i < VertexBuffer::MaxBuffers; i++) {
		if (m_Input[i] != NULL) {
			m_Input[i]->unlock();
		}

		if (m_Output[i] != NULL) {
			m_Output[i]->unlock();
		}
	}

	return false;
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
void VertexConverter::addAttributes(const VertexBuffer* buf, unsigned int idx, VertexConverter::Attribute* attribs) {
	unsigned int components;
	unsigned int i, j;
	unsigned int attributes;
	unsigned int type;
	unsigned int offset;
	const VertexFormat* fmt;
	const VertexFormat::Component* cmp;
	
	fmt = &buf->getFormat();
	components = fmt->getSize();

	for (i = 0; i < components; i++) {
		cmp = &fmt->at(i);
		attributes = cmp->getAttribute();
		if (attributes == 0)
			continue;
		type = cmp->getType();
		if (type == 0)
			continue;
		offset = cmp->getOffset();

		for (j = 0; j < Attributes::MaxAttributes; j++) {
			if ((attributes & (1 << j)) == 0)
				continue;
			attribs[j].m_Buffer = (VertexBuffer*)buf;
			attribs[j].m_BufferIdx = idx;
			attribs[j].m_Type = (VertexFormat::Type)type;
			attribs[j].m_Offset = offset;
			attribs[j].m_TypeBase = ((type >> Types::TypeShift) & Types::TypeMask);
			attribs[j].m_TypeSize = ((type >> Types::SizeShift) & Types::SizeMask);
			attribs[j].m_TypeCount = ((type >> Types::CountShift) & Types::CountMask);
		}
	}
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
void VertexConverter::removeAttributes(const VertexBuffer* buf, VertexConverter::Attribute* attribs) {
	for (unsigned int i = 0; i < Attributes::MaxAttributes; i++) {
		if (attribs[i].m_Buffer != buf)
			continue;
		attribs[i].m_Buffer = NULL;
		attribs[i].m_BufferIdx = 0;
		attribs[i].m_Type = (Types::Type)0;
		attribs[i].m_Offset = 0;
		attribs[i].m_TypeBase = 0;
		attribs[i].m_TypeSize = 0;
		attribs[i].m_TypeCount = 0;
	}
}


}; };
