/* 
  Magnum 
  Copyright (C) 2002-2004 Kaya Memisoglu. All Rights Reserved.

  This software is provided 'as-is', without any express or implied
  warranty.  In no event will the authors be held liable for any damages
  arising from the use of this software.

  Permission is granted to anyone to use this software for any purpose,
  including commercial applications, and to alter it and redistribute it
  freely, under the terms in LICENSE.TXT.

  Kaya Memisoglu  (k.memisoglu@dimajix.de)
*/

#include "gui.bitmap.h"
#include "gui.exception.h"

#include "base.string.inl"
#include "gui.bitmap.inl"
#include "graphics.point.inl"
#include "graphics.dimension.inl"
#include "graphics.rectangle.inl"
#include "graphics.image.inl"
#include "graphics.palette.inl"
#include "graphics.colour.inl"


#ifndef __FORWINDOWS__
#	error "This file is only for Windows"
#endif

namespace magnum { namespace gui {
using magnum::types::dword;

/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
bool Bitmap::createBitmap() {
	BITMAPINFO* bmpInfo;
	HBITMAP		oldBmp;
	RGBQUAD*  	palette;
	int			linesize,i;


	bmpInfo = (BITMAPINFO *) new byte[sizeof(BITMAPINFO)+256*sizeof(RGBQUAD)];
	if (bmpInfo == NULL)
		return false;
	palette = &bmpInfo->bmiColors[0];

	bmpInfo->bmiHeader.biSize = sizeof(BITMAPINFO) - sizeof(RGBQUAD);
	bmpInfo->bmiHeader.biWidth = getWidth();
	bmpInfo->bmiHeader.biHeight = -getHeight();
	bmpInfo->bmiHeader.biPlanes = 1;
	bmpInfo->bmiHeader.biXPelsPerMeter = 0;
	bmpInfo->bmiHeader.biYPelsPerMeter = 0;
	bmpInfo->bmiHeader.biClrUsed = 0;
	bmpInfo->bmiHeader.biClrImportant = 0;

	switch (getType()) {
		case BGRA8:
			bmpInfo->bmiHeader.biBitCount = 32;
			bmpInfo->bmiHeader.biCompression = BI_RGB;
			bmpInfo->bmiHeader.biSizeImage = getHeight() * getWidth() * 4;
			linesize = getWidth() * 4;
			break;
		case BGR8:
			bmpInfo->bmiHeader.biBitCount = 24;
			bmpInfo->bmiHeader.biCompression = BI_RGB;
			bmpInfo->bmiHeader.biSizeImage = getHeight() * getWidth() * 3;
			linesize = getWidth() * 3;
			break;
		case RGB5:
			bmpInfo->bmiHeader.biBitCount = 16;
			bmpInfo->bmiHeader.biCompression = BI_BITFIELDS;
			bmpInfo->bmiHeader.biSizeImage = getHeight() * getWidth() * 2;
			linesize = getWidth() * 2;
			((dword *)palette)[0] = 0x7c00;
			((dword *)palette)[1] = 0x03e0;
			((dword *)palette)[2] = 0x001f;
			break;
		case R5G6B5:
			bmpInfo->bmiHeader.biBitCount = 16;
			bmpInfo->bmiHeader.biCompression = BI_BITFIELDS;
			bmpInfo->bmiHeader.biSizeImage = getHeight() * getWidth() * 2;
			linesize = getWidth() * 2;
			((dword *)palette)[0] = 0xf800;
			((dword *)palette)[1] = 0x07e0;
			((dword *)palette)[2] = 0x001f;
			break;
		case GR8:
			bmpInfo->bmiHeader.biBitCount = 8;
			bmpInfo->bmiHeader.biCompression = BI_RGB;
			bmpInfo->bmiHeader.biSizeImage = getHeight() * getWidth();
			linesize = getWidth();
			for (i=0; i<256; i++) {
				palette[i].rgbRed=
				palette[i].rgbGreen=
				palette[i].rgbBlue=(BYTE)i;
				palette[i].rgbReserved=0;
			}
			break;
		case Index8:
			bmpInfo->bmiHeader.biBitCount = 8;
			bmpInfo->bmiHeader.biCompression = BI_RGB;
			bmpInfo->bmiHeader.biSizeImage = getHeight() * getWidth();
			linesize = getWidth();
			for (i=0; i<256; i++) {
				palette[i].rgbRed=
				palette[i].rgbGreen=
				palette[i].rgbBlue=(BYTE)i;
				palette[i].rgbReserved=0;
			}
			break;
		default:
			delete bmpInfo;
			return false;
	}

	m_Device = CreateCompatibleDC(NULL);
	if (m_Device == NULL) {
		delete[] (byte *)bmpInfo;
		return false;
	}

	m_Bitmap = CreateDIBSection( m_Device, bmpInfo, DIB_RGB_COLORS, (void **)&m_Pixels, NULL, NULL);
	if (m_Bitmap == NULL) {
		delete[] (byte *)bmpInfo;
		ReleaseDC(NULL, m_Device);
		m_Device = NULL;
		return false;
	}

	oldBmp = (HBITMAP)SelectObject(m_Device, m_Bitmap);
	if (oldBmp == NULL) {
		delete[] (byte *)bmpInfo;
		ReleaseDC(NULL, m_Device);
		DeleteObject(m_Bitmap);
		m_Device = NULL;
		m_Bitmap = NULL;
		return false;
	}
	DeleteObject(oldBmp);

	setLinesize(linesize);

	return true;
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
bool Bitmap::releaseBitmap() {
	if (m_Bitmap != NULL)
		DeleteObject(m_Bitmap);
	if (m_Device != NULL)
		ReleaseDC(NULL, m_Device);
	return true;
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
bool Bitmap::adjustBitmap() {
	switch (getType()) {
		case RGBA8:
			setType(BGRA8);
			return true;
		case RGB8:
			setType(BGR8);
			return true;
		case GR8:
		case BGRA8:
		case BGR8:
		case RGB5:
		case R5G6B5:
		case Index8:
			return true;
		default:
			return false;
	}
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
void* Bitmap::onLock() {
	return m_Pixels;
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
bool Bitmap::onGetPalette(Palette<byte>* pal) const {
	using magnum::graphics::Colour;

	if (getType() != Image::Index8)
		return false;

	RGBQUAD *color = new RGBQUAD[256];
	int i;

	GetDIBColorTable(m_Device, 0, 256, color);

	for (i = 0; i < 256; i++) {
		pal->setColour(Colour<byte>(color[i].rgbRed, color[i].rgbGreen, color[i].rgbBlue), i);
	}

	delete[] color;

	return true;
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
bool Bitmap::onSetPalette(const Palette<byte>& pal) {
	if (getType() != Image::Index8)
		return false;

	RGBQUAD *color = new RGBQUAD[256];
	int i;

	for (i = 0; i < 256; i++) {
		color[i].rgbRed = pal.getColour(i).R;
		color[i].rgbGreen = pal.getColour(i).G;
		color[i].rgbBlue = pal.getColour(i).B;
		color[i].rgbReserved = 0;
	}

	SetDIBColorTable(m_Device, 0, 256, color);
	delete[] color;

	return true;
}


}; };
