/* 
  Magnum 
  Copyright (C) 2002-2004 Kaya Memisoglu. All Rights Reserved.

  This software is provided 'as-is', without any express or implied
  warranty.  In no event will the authors be held liable for any damages
  arising from the use of this software.

  Permission is granted to anyone to use this software for any purpose,
  including commercial applications, and to alter it and redistribute it
  freely, under the terms in LICENSE.TXT.

  Kaya Memisoglu  (k.memisoglu@dimajix.de)
*/

#include "gui.opengl.h"
#include "gui.window.h"
#include "gui.context.h"
#include "gui.bitmap.h"
#include "container.chain.h"
#include "graphics.dimension.h"
#include "graphics.rectangle.h"
#include "system.h"
#include "application.message.h"
#include "graphics3d.context.h"
#include "graphics3d.vertexformat.h"

#include "base.string.inl"
#include "object.class.inl"
#include "object.inl"
#include "gui.opengl.inl"
#include "gui.view.inl"
#include "gui.window.inl"
#include "gui.context.inl"
#include "container.chain.inl"
#include "container.iterator.inl"
#include "graphics.transform.inl"
#include "graphics.dimension.inl"
#include "graphics.rectangle.inl"
#include "graphics.point.inl"
#include "graphics3d.context.inl"

#include <GL/gl.h>

#ifndef __FORWINDOWS__
#error "This file is only for Windows"
#endif

namespace magnum { namespace gui {


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
int OpenGL::chooseFormat(HDC deviceHandle, PIXELFORMATDESCRIPTOR* pfd) const {
	int pf;

    memset(pfd, 0, sizeof(PIXELFORMATDESCRIPTOR));
    pfd->nSize       = sizeof(PIXELFORMATDESCRIPTOR);
    pfd->nVersion    = 1;
    pfd->dwFlags     = PFD_DRAW_TO_WINDOW |PFD_SUPPORT_OPENGL | PFD_DOUBLEBUFFER;
    pfd->iPixelType  = PFD_TYPE_RGBA;
    pfd->cColorBits  = m_ColourBits;
	pfd->cAlphaBits	= m_AlphaBits;
	pfd->cAccumBits  = m_AccumulationBits;
	pfd->cDepthBits  = m_DepthBits;
	pfd->cStencilBits= m_StencilBits;
	pfd->cAuxBuffers = m_AuxBits;
	pfd->iLayerType  = PFD_MAIN_PLANE; 

	pf = ChoosePixelFormat(deviceHandle, pfd);
	if (pf == 0)
		return 0;

	if (SetPixelFormat(deviceHandle, pf, pfd) == FALSE)
		return 0;

	return pf;
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
bool OpenGL::adjustFormat(HDC deviceHandle, PIXELFORMATDESCRIPTOR* pfd) {
	int pf;

	pf = chooseFormat(deviceHandle, pfd);
	if (pf == 0) 
		return false;

	// Now get the real pixel format of the device
	DescribePixelFormat(deviceHandle, pf, sizeof(PIXELFORMATDESCRIPTOR), pfd);
    m_ColourBits = pfd->cColorBits;
	m_AlphaBits = pfd->cAlphaBits;
	m_AccumulationBits = pfd->cAccumBits;
	m_DepthBits = pfd->cDepthBits;
	m_StencilBits = pfd->cStencilBits;
	m_AuxBits = pfd->cAuxBuffers;
	return true;
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
bool OpenGL::verifyFormat() const {
	PIXELFORMATDESCRIPTOR pfd;
	HDC deviceHandle;
	int pf;

	deviceHandle = CreateDC("DISPLAY", NULL, NULL, NULL);

	pf = chooseFormat(deviceHandle, &pfd);
	if (pf == 0) {
		DeleteDC(deviceHandle);
		return false;
	}

	DeleteDC(deviceHandle);
	return true;
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
bool OpenGL::adjustFormat() {
	PIXELFORMATDESCRIPTOR pfd;
	HDC deviceHandle;
	bool ret;

	deviceHandle = CreateDC("DISPLAY", NULL, NULL, NULL);

	ret = adjustFormat(deviceHandle, &pfd);

	DeleteDC(deviceHandle);
	return ret;
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
bool OpenGL::onAttachWindow() {
	View::onAttachWindow();
	
	int n;
	PIXELFORMATDESCRIPTOR pfd;
	LOGPALETTE* lpPal;
	
	// Get device handle
	HDC deviceHandle = View::getContext()->getDeviceHandle();
	
	// Get WIndow
	HWND windowHandle = View::getContext()->getWindowHandle();
	
	SetWindowLongPtr(windowHandle, GWL_WNDPROC, (LONG_PTR)windowProc);
	

	// Adjust requested format
	if (adjustFormat(deviceHandle, &pfd) == false)
		return false;

	// Check if we need a palette
	if (pfd.dwFlags & PFD_NEED_PALETTE ||
		pfd.iPixelType == PFD_TYPE_COLORINDEX) {

		n = 1 << pfd.cColorBits;
		if (n > 256) 
			n = 256;

		lpPal = (LOGPALETTE*)new char[sizeof(LOGPALETTE) +
						sizeof(PALETTEENTRY) * n];
		memset(lpPal, 0, sizeof(LOGPALETTE) + sizeof(PALETTEENTRY) * n);
		lpPal->palVersion = 0x300;
		lpPal->palNumEntries = n;

		GetSystemPaletteEntries(deviceHandle, 0, n, &lpPal->palPalEntry[0]);
    
		/* if the pixel type is RGBA, then we want to make an RGB ramp,
		   otherwise (color index) set individual colors. */
		if (pfd.iPixelType == PFD_TYPE_RGBA) {
			int redMask = (1 << pfd.cRedBits) - 1;
			int greenMask = (1 << pfd.cGreenBits) - 1;
			int blueMask = (1 << pfd.cBlueBits) - 1;
			int i;

			/* fill in the entries with an RGB color ramp. */
			for (i = 0; i < n; ++i) {
			lpPal->palPalEntry[i].peRed = 
				(((i >> pfd.cRedShift)   & redMask)   * 255) / redMask;
			lpPal->palPalEntry[i].peGreen = 
				(((i >> pfd.cGreenShift) & greenMask) * 255) / greenMask;
			lpPal->palPalEntry[i].peBlue = 
				(((i >> pfd.cBlueShift)  & blueMask)  * 255) / blueMask;
			lpPal->palPalEntry[i].peFlags = 0;
			}
		} else {
			lpPal->palPalEntry[0].peRed = 0;
			lpPal->palPalEntry[0].peGreen = 0;
			lpPal->palPalEntry[0].peBlue = 0;
			lpPal->palPalEntry[0].peFlags = PC_NOCOLLAPSE;
			lpPal->palPalEntry[1].peRed = 255;
			lpPal->palPalEntry[1].peGreen = 0;
			lpPal->palPalEntry[1].peBlue = 0;
			lpPal->palPalEntry[1].peFlags = PC_NOCOLLAPSE;
			lpPal->palPalEntry[2].peRed = 0;
			lpPal->palPalEntry[2].peGreen = 255;
			lpPal->palPalEntry[2].peBlue = 0;
			lpPal->palPalEntry[2].peFlags = PC_NOCOLLAPSE;
			lpPal->palPalEntry[3].peRed = 0;
			lpPal->palPalEntry[3].peGreen = 0;
			lpPal->palPalEntry[3].peBlue = 255;
			lpPal->palPalEntry[3].peFlags = PC_NOCOLLAPSE;
		}

		m_PaletteHandle = CreatePalette(lpPal);
		if (m_PaletteHandle) {
			SelectPalette(deviceHandle, m_PaletteHandle, FALSE);
			RealizePalette(deviceHandle);
		}

		delete (char*)lpPal;
    }


	m_OpenGLContext = wglCreateContext(deviceHandle);
	wglMakeCurrent(deviceHandle, m_OpenGLContext);

	// Initialize Standard Viewport
	Rectangle<float> bounds = convertToWindow(getBounds());
	glViewport(0, 0, (int)bounds.getWidth(), (int)bounds.getHeight());

	m_Context = magnum::graphics3d::Context::getCurrent();

	return true;
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
bool OpenGL::onDetachWindow() {
	if (m_Context != NULL) {	
		wglMakeCurrent(NULL, NULL);
		wglDeleteContext(m_OpenGLContext); 
		m_OpenGLContext = NULL;
		m_Context = NULL;
	}
	
	View::onDetachWindow();

	return true;
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
LRESULT CALLBACK OpenGL::windowProc(HWND hwnd, UINT Message,WPARAM wParam, LPARAM lParam) {
	OpenGL* view = (OpenGL *)(LONG_PTR)GetWindowLongPtr(hwnd,0);

	switch(Message) {
		case WM_PALETTECHANGED:
PRINT("OpenGL::windowProc: WM_PALETTECHANGED");
			if (hwnd == (HWND)wParam)
				break;
			/* fall through to WM_QUERYNEWPALETTE */
		case WM_QUERYNEWPALETTE:
PRINT("OpenGL::windowProc: WM_QUERYNEWPALETTE");
			if (view->m_PaletteHandle) {
				UnrealizeObject(view->m_PaletteHandle);
				SelectPalette(view->View::getContext()->getDeviceHandle(), view->m_PaletteHandle, FALSE);
				RealizePalette(view->View::getContext()->getDeviceHandle());
				return TRUE;
			}
			return FALSE;
		case WM_PAINT:
PRINT("OpenGL::windowProc: WM_PAINT");
			ValidateRect(hwnd,NULL);
			if (view->m_Context == NULL)
				return 0;
			view->m_Context->makeCurrent();
			view->onRender();
			SwapBuffers(view->getDeviceHandle());
			return 0;
	}

	return View::windowProc(hwnd, Message, wParam, lParam);
}


}; };
