/* 
  Magnum 
  Copyright (C) 2002-2004 Kaya Memisoglu. All Rights Reserved.

  This software is provided 'as-is', without any express or implied
  warranty.  In no event will the authors be held liable for any damages
  arising from the use of this software.

  Permission is granted to anyone to use this software for any purpose,
  including commercial applications, and to alter it and redistribute it
  freely, under the terms in LICENSE.TXT.

  Kaya Memisoglu  (k.memisoglu@dimajix.de)
*/

#include "media.node.h"
#include "media.device.h"
#include "media.pin.h"
#include "media.graph.h"
#include "media.buffer.h"

#include "base.string.inl"
#include "object.class.inl"
#include "object.inl"
#include "system.thread.inl"
#include "system.signal.inl"
#include "system.mutex.inl"
#include "container.list.inl"
#include "container.chain.inl"
#include "container.queue.inl"
#include "container.iterator.inl"
#include "media.node.inl"
#include "media.device.inl"
#include "media.handler.inl"
#include "media.pin.inl"
#include "media.graph.inl"
#include "media.buffer.inl"


namespace magnum { namespace media {

const char Node::DefaultInput[] = "DefaultInput";
const char Node::DefaultOutput[] = "DefaultOutput";
const char Node::DefaultFilter[] = "DefaultFilter";
const char Node::DefaultNode[] = "DefaultNode";


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
Node::Node(Device* dev, const char* name, Type t) {
	m_Name = name;
	m_Graph = NULL;
	m_Device = dev;
	m_State = Stopped;
	m_Type = t;

	m_Handlers.insert(this);
	release();

	if (m_Device != NULL) {
		m_Device->refer();
	}
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
Node::~Node() {
	refer();
	m_Handlers.remove(this);

	if (m_Device != NULL) {
		m_Device->release();
	}
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
Buffer* Node::receiveBuffer(InputPin* pin,bool block) {
	if (!pin->isConnected()) {
		return NULL;
	}

	OutputPin* cpin = (OutputPin *)pin->getConnection();
	Node* cnode = cpin->getNode();
	Buffer* buffer;

	// Check queue of pin. Empty queue->request buffer
	pin->m_QueueMutex.lock();
	buffer = pin->m_BufferQueue.removeHead();
	pin->m_QueueSignal.reset();
	pin->m_QueueMutex.unlock();

	// Did we really get a buffer?
	if (buffer == NULL) {
		// inform connected node that we need a buffer
		Message msg(Message::RequestBuffer);
		msg.setPin(cpin);
		cnode->sendMessage(msg);

		// No blocking?
		if (block == false)
			return NULL;

		// wait for buffer
		pin->m_QueueSignal.wait();
		pin->m_QueueMutex.lock();
		buffer = pin->m_BufferQueue.removeHead();
		pin->m_QueueMutex.unlock();
	}

	return buffer;
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
bool Node::sendBuffer(OutputPin* pin,Buffer* buffer,bool block) {
	if (!pin->isConnected()) {
		buffer->release();
		return true;	
	}

	InputPin* cpin = (InputPin *)pin->getConnection();
	Node* cnode = cpin->getNode();

	buffer->m_Producer = pin;
	buffer->m_Consumer = cpin;

	// fire handlers
	Message msg;
	msg.setCommand(Message::SendBuffer);
	msg.setPin(pin);
	msg.setPointer(buffer);
	sendMessage(msg);

	msg.setCommand(Message::ReceiveBuffer);
	msg.setPin(cpin);
	cnode->sendMessage(msg);

	cpin->m_QueueMutex.lock();
	cpin->m_BufferQueue.insert(buffer);
	cpin->m_QueueMutex.unlock();
	cpin->m_QueueSignal.raise();

	return true;
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
bool Node::publishOutputPin(OutputPin* pin,bool def) {
	PinInfo info;
	info.m_Name = pin->getName();
	info.m_Pin = pin;
	if (def)
		m_DefaultOutput = pin;
	m_OutputPins.insert(info);

	return true;
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
bool Node::publishInputPin(InputPin* pin,bool def) {
	PinInfo info;
	info.m_Name = pin->getName();
	info.m_Pin = pin;
	if (def)
		m_DefaultInput = pin;
	m_InputPins.insert(info);

	return true;
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
InputPin* Node::getInputPin(const char* which) {
	InputPin* pin;

	if (which == NULL) {
		pin = m_DefaultInput;
	} else if (strcmp(which, Pin::DefaultInput) == 0) {
		pin = m_DefaultInput;
	} else {
		pin = NULL;
	}

	return pin;
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
OutputPin* Node::getOutputPin(const char* which) {
	OutputPin* pin;

	if (which == NULL) {
		pin = m_DefaultOutput;
	} else if (strcmp(which, Pin::DefaultOutput) == 0) {
		pin = m_DefaultOutput;
	} else {
		pin = NULL;
	}

	return pin;
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
bool Node::sendMessage(const Message& msg) {
	Message message(msg);
	message.m_Node = this;
	Handler* handler;
	Chain<Handler>::Iterator iter;

	for (iter = m_Handlers.begin(); iter != m_Handlers.end(); iter++) {
		handler = &(*iter);
		handler->handleMessage(message);
	}

	return true;
}


}; };
