#include "system/thread.h"
#include "media/path.h"
#include "media/node.h"
#include "media/pin.h"
#include "media/buffer.h"

#include "base/string.inl"
#include "object/class.inl"
#include "container/iterator.inl"
#include "container/list.inl"
#include "container/chain.inl"
#include "object/object.inl"
#include "system/thread.inl"
#include "media/path.inl"
#include "media/pin.inl"
#include "media/node.inl"
#include "media/buffer.inl"
#include "media/handler.inl"


namespace magnum { namespace media {

/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
Path::Path() {
	m_State = Node::Stopped;
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
Path::~Path() {

}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
bool Path::addNode(Node* node) {
	if (m_State != Node::Stopped)
		return false;
	if (node->getPath() != NULL)
		return false;

	m_Nodes.insert(node);
	node->m_Path = this;
	return true;
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
bool Path::removeNode(Node* node) {
	if (m_State != Node::Stopped)
		return false;
	if (node->getPath() != this)
		return false;

	// Disconnect all pins

	m_Nodes.remove(node);
	node->m_Path = NULL;
	return true;
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
bool Path::connect(OutputPin* src, InputPin* dst) {
	if (m_State != Node::Stopped)
		return false;
	if (src->isConnected() || dst->isConnected())
		return false;
	if (src->getNode() == dst->getNode())
		return false;
	if ((src->getPath() != this) || (dst->getPath() != this))
		return false;

	// Connect both nodes
	src->m_Connection = dst;
	dst->m_Connection = src;
	src->getNode()->onConnect(src);
	dst->getNode()->onConnect(dst);

	// fire handlers
	Message msg(Message::Connect);
	msg.setPin(src);
	src->getNode()->sendMessage(msg);
	msg.setPin(dst);
	dst->getNode()->sendMessage(msg);

	return true;
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
bool Path::disconnect(OutputPin* src, InputPin* dst) {
	if (m_State != Node::Stopped)
		return false;
	if (src->getConnection() != dst)
		return false;
	if (dst->getConnection() != src)
		return false;
	if ((src->getPath() != this) || (dst->getPath() != this))
		return false;

	// Disconnect both nodes
	src->m_Connection = NULL;
	dst->m_Connection = NULL;
	src->getNode()->onDisconnect(src);
	dst->getNode()->onDisconnect(dst);

	// fire handlers
	Message msg(Message::Disconnect);
	msg.setPin(src);
	src->getNode()->sendMessage(msg);
	msg.setPin(dst);
	dst->getNode()->sendMessage(msg);

	return true;
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
bool Path::start() {
	Message msg(Message::Start);
	Node* node;
	Chain<Node>::Iterator iter = m_Nodes.begin();

	if (m_State != Node::Stopped)
		return false;

	while (iter != m_Nodes.end()) {
		node = (Node *)&iter;
		node->onStart();
		node->sendMessage(msg);
		iter++;
	}

	m_State = Node::Running;

	return true;
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
bool Path::stop() {
	Message msg(Message::Stop);
	Node* node;
	Chain<Node>::Iterator iter = m_Nodes.begin();

	if ((m_State != Node::Running) && (m_State != Node::Paused))
		return false;

	while (iter != m_Nodes.end()) {
		node = (Node *)&iter;
		node->onStop();
		node->sendMessage(msg);
		iter++;
	}

	m_State = Node::Stopped;

	return true;
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
bool Path::pause() {
	Message msg(Message::Pause);
	Node* node;
	Chain<Node>::Iterator iter = m_Nodes.begin();

	if (m_State != Node::Running)
		return false;

	while (iter != m_Nodes.end()) {
		node = (Node *)&iter;
		node->onPause();
		node->sendMessage(msg);
		iter++;
	}

	m_State = Node::Paused;

	return true;
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
bool Path::unpause() {
	Message msg(Message::Unpause);
	Node* node;
	Chain<Node>::Iterator iter = m_Nodes.begin();

	if (m_State != Node::Paused)
		return false;

	while (iter != m_Nodes.end()) {
		node = (Node *)&iter;
		node->onUnpause();
		node->sendMessage(msg);
		iter++;
	}

	m_State = Node::Running;

	return true;
}

}; };
