/* 
  Magnum 
  Copyright (C) 2002-2004 Kaya Memisoglu. All Rights Reserved.

  This software is provided 'as-is', without any express or implied
  warranty.  In no event will the authors be held liable for any damages
  arising from the use of this software.

  Permission is granted to anyone to use this software for any purpose,
  including commercial applications, and to alter it and redistribute it
  freely, under the terms in LICENSE.TXT.

  Kaya Memisoglu  (k.memisoglu@dimajix.de)
*/

#include "scene3d.node.h"
#include "scene3d.group.h"
#include "scene3d.root.h"
#include "scene3d.locale.h"
#include "scene3d.universe.h"

#include "scene3d.node.inl"
#include "scene3d.locale.inl"

#include "typeinfo.h"

namespace magnum { namespace scene3d {

/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
Node::Node(Node::Type t) {
	m_Parent = NULL;
	m_Next = m_Prev = NULL;
	m_NextDirty = m_PrevDirty = NULL;
	m_Type = t;
	m_Locale = NULL;
	m_Flags = Enabled | AutoBounds;
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
Node::~Node() {
	ASSERT(m_Parent == NULL);
	ASSERT(m_Next == NULL);
	ASSERT(m_Prev == NULL);
	dequeueUpdateTransform();
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
void Node::enable() {
	if (m_Flags & Enabled)
		return;
	if (m_Parent) {
		if (!m_Parent->testFlag(Enabled)) {
			m_Parent->enable();
		}
	}
	m_Flags |= Enabled;
	onEnable();
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
void Node::disable() {
	if ((m_Flags & Enabled) == 0)
		return;
	m_Flags &= ~Enabled;
	onDisable();
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
void Node::enqueueUpdateTransform() {
	m_Flags |= DirtyTransform;

	// m_PrevDirty is the indicator wether the node is already queued
	// for updating
	if (m_PrevDirty != NULL)
		return;

	// No parent - no updating needed (?)
	if (m_Parent == NULL)
		return;

	// No locale means that we do not have acces to a dirty queue
	if (m_Locale == NULL)
		return;

	Universe* universe = getUniverse();
	ASSERT(universe != NULL);
	Node* anchor;

	// Get suitable node to insert after
	if (m_Parent != NULL) {
		if (((Node*)m_Parent)->m_Flags & DirtyTransform)
			anchor = m_Parent;
		else
			anchor = &universe->m_DirtyNodes;
	} else {
		anchor = &universe->m_DirtyNodes;
	}

	// Link this node behind anchor
	m_NextDirty = anchor->m_NextDirty;
	m_PrevDirty = anchor;
	anchor->m_NextDirty = this;

	// Enqueue all children
	if (m_Type & GroupType) {
		Node* node = ((Group*)this)->m_First;
		while(node != NULL) {
			node->applyParentTransform();
			node = node->m_Next;
		}
	}
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
void Node::dequeueUpdateTransform() {
	// Is this node really enqueued?
	if (m_PrevDirty == NULL)
		return;

	// Parentless nodes won't be enqueued
	ASSERT(m_Parent != NULL);

	// Check if parent is enqueued...
	if (((Node*)m_Parent)->m_PrevDirty != NULL) {
		m_Flags |= DirtyTransform;
		return;
	}		

	m_Flags &= ~DirtyTransform;

	m_PrevDirty->m_NextDirty = m_NextDirty;
	if (m_NextDirty != NULL)
		m_NextDirty->m_PrevDirty = m_PrevDirty;

	m_NextDirty = m_PrevDirty = NULL;
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
Node* Node::getRoot() const {
	const Node* node = this;
	while (node->m_Parent != NULL) {
		node = node->m_Parent;
	}

	return (Node*)node;
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
void Node::onAttachLocale(Locale* l) {
	ASSERT(l != NULL);
	ASSERT(m_Locale == NULL);
	m_Locale = l;
	if (m_Parent != NULL) {
		if ((m_Parent->getType() & TransformType) || m_Parent->testFlag(ParentTransform))
			setFlag(ParentTransform);
	}
	enqueueUpdateTransform();
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
void Node::onDetachLocale(Locale* l) {
	ASSERT(l != NULL);
	ASSERT(m_Locale == l);
	clearFlag(ParentTransform);
	dequeueUpdateTransform();
	m_Locale = NULL;
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
void Node::onEnable() {
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
void Node::onDisable() {
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
void Node::onTransform(const Transform3<float>& ) {
}


}; };
